"""
This module contains the implementation of the data handling for the icomox.
Parsers and descriptor objects are implemented.
Technical details about the device can be found in the following document:
"""
# ============================ imports =========================================
import datetime
import logging
import struct
import time
from abc import abstractmethod, ABC
from enum import Enum
from common import iCOMOX_messages, messages_utils, helpers, common
from common.common import iCOMOXs
from common.iCOMOX_list import sCOMOX_TCPIP, sCOMOXs_list
from common.iCOMOX_messages import OUT_MSG_Hello, OUT_MSG_SetConfiguration, OUT_MSG_Reset
from txp.common.config import settings
import logging

log = logging.getLogger(__name__)
log.setLevel(settings.txp.general_log_level)


# ==============================================================================
# Responses descriptors implementation
# ==============================================================================

SOFTWARE_RESET_TYPE=0

class IcomoxResponsesTypes(Enum):
    HELLO_MESSAGE_RESPONSE = "HELLO_MESSAGE_RESPONSE"
    CONFIGURATION_RESPONSE = "CONFIGURATION_RESPONSE"
    EEPROM_RESPONSE = "EEPROM_RESPONSE"
    EEPROM_VALUES_RESPONSE = "EEPROM_VALUES_RESPONSE"
    OK_RESPONSE = "OK_RESPONSE"


class IcomoxAPIMessageDescriptor(ABC):
    def __init__(self, type):
        self._messageType = type

    @staticmethod
    @abstractmethod
    def parseFromRawMessage(msg):
        pass


class IcomoxEEPROMValueResponseDescriptor(IcomoxAPIMessageDescriptor):

    @staticmethod
    def parseFromRawMessage(msg):
        # TODO: implement the message parsing and builder of the class
        return IcomoxEEPROMValueResponseDescriptor()

    def __init__(self):
        IcomoxAPIMessageDescriptor.__init__(IcomoxResponsesTypes.EEPROM_VALUES_RESPONSE)


class IcomoxEEPROMStatusResponseDescriptor(IcomoxAPIMessageDescriptor):
    def __init__(self):
        IcomoxAPIMessageDescriptor.__init__(IcomoxResponsesTypes.EEPROM_RESPONSE)
        self._resultValue = None

    def _setResultValue(self, result):
        self._resultValue = result

    def getResultValue(self):
        return self._resultValue

    def isEEMPROMStateOK(self):
        return self._resultValue == 0

    @staticmethod
    def parseFromRawMessage(msg):
        count, address, result = struct.unpack_from("<BHB", msg, 1)
        resultValue = "EEPROM_OK"
        if result != 0:
            resultValue = "ERROR"
        descriptor = IcomoxEEPROMStatusResponseDescriptor()
        descriptor._setResultValue(resultValue)
        return descriptor


class IcomoxOKResponseDescriptor(IcomoxAPIMessageDescriptor):
    def __init__(self):
        IcomoxAPIMessageDescriptor.__init__(IcomoxResponsesTypes.OK_RESPONSE)
        self._responseValue = None

    def getResponseValue(self):
        return "OK"

    @staticmethod
    def parseFromRawMessage(msg):
        # TODO: evaluate the value to implement the parser. Evaluate if not OK response is sent.
        descriptor = IcomoxOKResponseDescriptor()
        return descriptor


class IcomoxCongurationResponseDescriptor(IcomoxAPIMessageDescriptor):
    def __init__(self):
        IcomoxAPIMessageDescriptor.__init__(IcomoxResponsesTypes.CONFIGURATION_RESPONSE)
        self._common = None
        self._repetition = None
        self._intervalInMinutes = None
        self._activeModules = None
        self._rawDataSensors = None
        self._maintenanceModuleSensors = None
        self._maintenanceModuleFlags = None
        self._maintenanceModuleAlpha = None

    def getResponseType(self):
        return self._messageType

    def _setCommon(self, common):
        self._common = common

    def getCommon(self):
        return self._common

    def _setRepetition(self, repetition):
        self._repetition = repetition

    def getRepetition(self):
        return self._repetition

    def _setIntervalInMinutes(self, intervalInMinutes):
        self._intervalInMinutes = intervalInMinutes

    def getIntervalInMinutes(self):
        return self._intervalInMinutes

    def _setActiveModules(self, activeModules):
        self._activeModules = activeModules

    def getActiveModules(self):
        return self._activeModules

    def _setRawDataSensors(self, rawDataSensors):
        self._rawDataSensors = rawDataSensors

    def getRawDataSensors(self):
        return self._rawDataSensors

    def _setMaintenanceModuleSensors(self, maintenanceModuleSensors):
        self._maintenanceModuleSensors = maintenanceModuleSensors

    def getMaintenanceModuleSensors(self):
        return self._maintenanceModuleSensors

    def _setMaintenanceModuleFlags(self, maintenanceModuleFlags):
        self._maintenanceModuleFlags = maintenanceModuleFlags

    def getMaintenanceModuleFlags(self):
        return self._maintenanceModuleFlags

    def _setMaintenanceModuleAlpha(self, maintenanceModuleAlpha):
        self._maintenanceModuleAlpha = maintenanceModuleAlpha

    def getMaintenanceModuleAlpha(self):
        return self._maintenanceModuleAlpha

    @staticmethod
    def parseFromRawMassage(self, msg):
        """
        This static method is the entry point to build the IcomoxCongurationResponseDescriptor instance.
        :param msg: bytearray containing the raw data of a configuration message response from icomox
        :return: IcomoxConfigurationResponseDescriptor
        """
        descriptor = IcomoxCongurationResponseDescriptor()

        Common, \
        Repetition, IntervalInMinutes, \
        ActiveModules, \
        RawDataSensors, \
        ML1_target, ML1_selSensor, ML1_countLearned, ML1_modelLearned, \
        MaintenanceModule_sensors, MaintenanceModule_flags, MaintenanceModule_alpha = iCOMOX_messages.IN_MSG_GetConfiguration(
            msg)

        descriptor._setCommon(Common)
        descriptor._setRepetition(Repetition)
        descriptor._setIntervalInMinutes(IntervalInMinutes)
        descriptor._setActiveModules(ActiveModules)
        descriptor._setRawDataSensors(RawDataSensors)
        descriptor._setMaintenanceModuleSensors(MaintenanceModule_sensors)
        descriptor._setMaintenanceModuleFlags(MaintenanceModule_flags)
        descriptor._setMaintenanceModuleAlpha(MaintenanceModule_alpha)

        return descriptor


class IcomoxHelloMessageDescriptor:
    def __init__(self):
        self._boardType = None
        self._boardVersion = None
        self._mcuSerialNumber = None
        self._firmwareReleaseVersion = None
        self._bitStatus = None

    def _setBitStatus(self, bitStatus):
        self._bitStatus = bitStatus

    def getBitStatus(self):
        return self._bitStatus

    def _setBoardType(self, boardType):
        self._boardType = boardType

    def getBoardType(self):
        return self._boardType

    def _setBoardVersion(self, boardVersionMajor, boardVertionMinor):
        self._boardVersion = messages_utils.iCOMOX_BoardVersion_to_Str(board_version_major=boardVersionMajor,
                                                                       board_version_minor=boardVertionMinor)

    def getBoardVersion(self):
        return self._boardVersion

    def _setMcuSerialNumber(self, mcuSerialNumber):
        self._mcuSerialNumber = mcuSerialNumber

    def getMcuSerialNumber(self):
        return self._mcuSerialNumber

    def getMcuSerialNumberString(self):
        return helpers.u8s_to_str(arr=self._mcuSerialNumber, separator="")

    def _setFirmwareReleaseVersion(self, hello_firmware_release_version_major, hello_firmware_release_version_minor,
                                   hello_firmware_release_version_patch):
        self._firmwareReleaseVersion = (hello_firmware_release_version_major << 16) + (
                hello_firmware_release_version_minor << 8) + hello_firmware_release_version_patch

    def getFirmwareReleaseVersion(self):
        return self._firmwareReleaseVersion

    @staticmethod
    def parseFromRawData(msg):
        """
        This static method is the entry point to build the IcomoxHelloMessageDescriptor instance.
        :param msg: bytearray containing the raw data of a configuration message response from icomox
        :return: IcomoxHelloMessageDescriptor
        """
        hello_board_type, \
        hello_board_version_major, hello_board_version_minor, \
        hello_mcu_serial_number, \
        hello_firmware_release_version_major, hello_firmware_release_version_minor, hello_firmware_release_version_patch, hello_firmware_release_version_branch, \
        hello_firmware_build_version_date_year, hello_firmware_build_version_date_month, hello_firmware_build_version_date_day, \
        hello_firmware_build_version_time_hour, hello_firmware_build_version_time_minute, hello_firmware_build_version_time_second, \
        hello_bit_status, \
        hello_product_part_number, hello_production_serial_number, hello_name, \
        smip_swMajor, smip_swMinor, smip_swPatch, smip_swBuild = iCOMOX_messages.IN_MSG_Hello(msg=msg)

        descriptor = IcomoxHelloMessageDescriptor()
        descriptor._setFirmwareReleaseVersion(hello_firmware_release_version_major,
                                              hello_firmware_release_version_minor,
                                              hello_firmware_release_version_patch)
        descriptor._setBoardType(hello_board_type)
        descriptor._setBitStatus(hello_bit_status)
        descriptor._setBoardVersion(hello_board_version_major, hello_board_version_minor)
        descriptor._setMcuSerialNumber(hello_mcu_serial_number)

        return descriptor


# ==============================================================================
# Module Functions
# ==============================================================================
def sendGetHelloMessageCommand(icomox: sCOMOX_TCPIP):
    """
    :param icomox: icomox tcp instance
    """
    if icomox is not None:
        icomox.transmit_buffer = OUT_MSG_Hello()
    else:
        raise Exception("Not valid icomox instance")

def sendSetConfigurationCommand(uniqueId,enableTransmition):
    iComox = iCOMOXs.find_by_UniuqeID(uniqueId)
    if iComox is not None:
        ConfigBitmask = 3
        ConfigModulesBitmask = 1
        LocalTimestamp = int(time.time() + (datetime.datetime.now() - datetime.datetime.utcnow()).total_seconds())   # mktime(time.localtime()))
        IntervalInMinutes=0
        Repetition=0
        if enableTransmition:
            Common = 65
        else:
            Common = 0
        ActiveModules = 1
        RawData_Sensors = 31
        # Module_AnomalyDetection_Command, Module_AnomalyDetection_stateToTrain, Module_AnomalyDetection_Sensors = self.WindowToModuleAnomalyDetection()
        Module_AnomalyDetection_Sensors = iCOMOX_messages.cCOMOX_SENSOR_BITMASK_ADXL356
        Module_AnomalyDetection_stateToTrain = 0
        Module_AnomalyDetection_Command = iCOMOX_messages.cANOMALY_DETECTION_COMMAND_Reset

        iComoxMessage = OUT_MSG_SetConfiguration(ConfigBitmask=ConfigBitmask, ConfigModulesBitmask=ConfigModulesBitmask, \
                                                        LocalTimestamp=LocalTimestamp, \
                                                        Common=Common, \
                                                        Repetition=Repetition, IntervalInMinutes=IntervalInMinutes, \
                                                        ActiveModules=ActiveModules, \
                                                        RawData_Sensors=RawData_Sensors, \
                                                        AnomalyDetection_Command=Module_AnomalyDetection_Command, AnomalyDetection_Sensors=Module_AnomalyDetection_Sensors, AnomalyDetection_StateToTrain=Module_AnomalyDetection_stateToTrain)


        iComox.transmit_buffer = iComoxMessage
    else:
        raise Exception("Not valid icomox instance")

def resetIcomox(uniqueId):
    iComox = iCOMOXs.find_by_UniuqeID(uniqueId)
    if iComox is not None:
        log.info("{} icomox is being reseting".format(uniqueId))
        iComox.transmit_buffer = OUT_MSG_Reset(SOFTWARE_RESET_TYPE)
    else:
        log.error("{} icomox reset attemp error".format(uniqueId))

def getHelloMessage(msg, icomox):
    """
    This module funtion tries to get the hello message object to identify the connected icomox EDGE

    :param msg: raw data from icomox
    :param icomox: icomox TCP type object
    :return: IcomoxHelloMessageDescriptor if message contains hello info. None if message is not hello type
    """
    if msg[0] == iCOMOX_messages.cCOMOX_MSG_CODE_Hello:
        return IcomoxHelloMessageDescriptor.parseFromRawData(msg)
    else:
        return None

def processSDKCommandResponse(msg):
    """
    this funtion module takes care of processing an icomox message containing a response from the SDK available commands API
    :param msg: bytearray() raw data from icomox
    :return: IcomoxAPIMessageDescriptor.  None if message is not valid.
    """
    msg_code = msg[0]
    response = None
    if msg_code == iCOMOX_messages.cCOMOX_MSG_CODE_Reset:
        response = IcomoxOKResponseDescriptor.parseFromRawMessage(msg)
    elif msg_code == iCOMOX_messages.cCOMOX_MSG_CODE_GetConfiguration:
        response = IcomoxCongurationResponseDescriptor.parseFromRawMessage(msg)
    elif msg_code == iCOMOX_messages.cCOMOX_MSG_CODE_SetConfiguration:
        response = IcomoxOKResponseDescriptor.parseFromRawMessage(msg)
    elif msg_code == iCOMOX_messages.cCOMOX_MSG_CODE_ReadEEPROM:
        response = IcomoxEEPROMValueResponseDescriptor.parseFromRawMessage(msg)
    elif msg_code == iCOMOX_messages.cCOMOX_MSG_CODE_WriteEEPROM or iCOMOX_messages.cCOMOX_MSG_CODE_VerifyEEPROM:
        response = IcomoxEEPROMStatusResponseDescriptor.parseFromRawMessage(msg)
    elif msg_code == iCOMOX_messages.cCOMOX_MSG_CODE_Debug:
        log.error("Unsupported message response implementation for cCOMOX_MSG_CODE_Debug type")
        response = IcomoxOKResponseDescriptor.parseFromRawMessage(msg)
    else:
        # TODO: review the possibility to raise an exception for getting invalid message.
        log.error("Non supported message code for SDK. Message code received {}".format(msg_code))
    return response


class IcomoxMessageType(Enum):
    REPORT_MESSAGE = "REPORT_MESSAGE"
    API_MESSAGE = "API_MESSAGE"
    HELLO = "HELLO"
    INVALID = "INVALID"

def getIcomoxMessageType(msg):
    msg_code = msg[0]
    messageType = IcomoxMessageType.INVALID

    if msg_code == iCOMOX_messages.cCOMOX_MSG_CODE_Hello:
        messageType = IcomoxMessageType.HELLO
    elif msg_code == iCOMOX_messages.cCOMOX_MSG_CODE_Report:
        messageType = IcomoxMessageType.REPORT_MESSAGE
    elif (msg_code & 7) > 0: #this is a validation forced by SDK the supported messages are in that range.
        messageType = IcomoxMessageType.API_MESSAGE

    return messageType

class IcomoxSensorMassageType(Enum):
    VIBRATION_SENSOR_ADXL362 = "VIBRATION_SENSOR_ADXL362"
    VIBRATION_SENSOR_ADXL356 = "VIBRATION_SENSOR_ADXL356"
    MAGNETOMETER_SENSOR_BMM150 = "MAGNETOMETER_SENSOR_BMM150"
    TEMPERATURE_SENSOR_ADT7410 = "TEMPERATURE_SENSOR_ADT7410"
    MICROPHONE_SENSOR_IM69D130 = "MICROPHONE_SENSOR_IM69D130"
    INVALID_SENSOR = "INVALID_SENSOR"

def getSignalSamples(msg):
    module, sensor, axis, timestamp = iCOMOX_messages.IN_MSG_Report(msg=msg)
    payload = msg[10:]
    samples = None
    sensorType = IcomoxSensorMassageType.INVALID_SENSOR
    if module == iCOMOX_messages.cMODULE_RawData:
        if iCOMOX_messages.cCOMOX_SENSOR_ADXL362 == sensor:
            samples = getVibrationSignalSamplesFromADXL362(payload)
            sensorType = IcomoxSensorMassageType.VIBRATION_SENSOR_ADXL362
        elif iCOMOX_messages.cCOMOX_SENSOR_ADXL356 == sensor:
            acc_x_units_g, acc_y_units_g, acc_z_units_g = common.ADXL356.msg_bytes_to_samples_g(payload=payload)
            samples = list([acc_x_units_g, acc_y_units_g, acc_z_units_g])
            sensorType =  IcomoxSensorMassageType.VIBRATION_SENSOR_ADXL356
        elif iCOMOX_messages.cCOMOX_SENSOR_BMM150 == sensor:
            mag_x_data, mag_y_data, mag_z_data = common.BMM150.msg_bytes_to_samples_uTesla(payload=payload)
            samples = list([mag_x_data, mag_y_data, mag_z_data])
            sensorType = IcomoxSensorMassageType.MAGNETOMETER_SENSOR_BMM150
        elif iCOMOX_messages.cCOMOX_SENSOR_ADT7410 == sensor:
            temp_data = common.ADT7410.msg_bytes_to_sample_Celsius(payload=payload)
            samples = [temp_data]
            sensorType = IcomoxSensorMassageType.TEMPERATURE_SENSOR_ADT7410
        elif iCOMOX_messages.cCOMOX_SENSOR_IM69D130 == sensor:
            samples_SPL = common.IM69D130.msg_bytes_to_samples_SPL(payload=payload)
            samples = list([samples_SPL])
            sensorType = IcomoxSensorMassageType.MICROPHONE_SENSOR_IM69D130

    return sensorType,samples

def getVibrationSignalSamplesFromADXL362(payload):
    acc_x_units_g, acc_y_units_g, acc_z_units_g = common.ADXL362.msg_bytes_to_samples_g(payload=payload)
    return list([acc_x_units_g, acc_y_units_g, acc_z_units_g])



"""
This module implements the Authentication view for the application.
"""
# ============================ imports =========================================
import oauthlib
import txp.cloud.controllers.authentication as auth
from txp.web.core_components.app_component import AppComponent
import logging
import streamlit as st

# TODO: This might be a problem. The logging configuration is being taken from `txp` package
from txp.common.config import settings

log = logging.getLogger(__name__)
log.setLevel(settings.txp.general_log_level)


class AuthenticationComponent(AppComponent):
    def __init__(self):
        super(AuthenticationComponent, self).__init__(
            component_key="authentication_view"
        )
        #  The redirection flow must be set manually. Streamlit doesn't provide a way to obtain the
        #  url host.
        self._CLIENT_OAUTH_SECRET = st.secrets["oauth_secrets"]
        self._FIREBASE_API_KEY = st.secrets["firebase_api_key"]
        self._GOOGLE_OAUTH_REDIRECT = st.secrets["oauth_redirect_url"]
        self._FIREBASE_DEFAULT_SERVICE_ACCOUNT = st.secrets[
            "firebase_default_service_account"
        ]

    def _render(self) -> None:
        if not self._is_google_auth_redirection():
            st.title("Ingresar a Tranxpert")
            google_oauth2_flow = auth.get_google_oauth_flow(
                self._CLIENT_OAUTH_SECRET, self._GOOGLE_OAUTH_REDIRECT
            )

            auth_url, _ = google_oauth2_flow.authorization_url()

            st.write(
                f"""
                    <h4>
                        <a target="_blank" href="{auth_url}" onclick="close();return false;">
                        Ingresar con Google
                        </a>
                    </h4>
                """,
                unsafe_allow_html=True,
            )
        else:
            log.info("User no authenticated. Starting authentication flow")
            self._authenticate_user_and_get_roles()

    @staticmethod
    def _is_google_auth_redirection() -> bool:
        """Returns true if there are query parameters in the URL that matches
        the redirected parameters by Google Auth flow after consent.

        More info:
            https://developers.google.com/identity/protocols/oauth2/web-server#handlingresponse
        """
        params = st.experimental_get_query_params()
        return "code" in params

    @staticmethod
    def _get_auth_code() -> str:
        """Returns the OAuthentication code from the URL Query params"""
        code = st.experimental_get_query_params()["code"][0]
        return code

    def _authenticate_user_and_get_roles(self):
        try:
            identity_platform_credentials = auth.sing_in_user_in_identity_platform(
                self._get_auth_code(),
                self._FIREBASE_API_KEY,
                self._CLIENT_OAUTH_SECRET,
                self._GOOGLE_OAUTH_REDIRECT,
            )
            # st.experimental_set_query_params(**{})
        except oauthlib.oauth2.rfc6749.errors.InvalidGrantError:
            st.experimental_set_query_params(**{})
            st.experimental_rerun()

        if not identity_platform_credentials:
            st.error(
                "Hubo un error al realizar la autenticación."
                "Intente nuevamente o contacte a soporte."
            )
            log.error("Could not sign-in user in Identity Platform")
            return

        self.app_state.authentication_info.identity_platform_credentials = (
            identity_platform_credentials
        )

        role_credentials, role_enum = auth.get_principal_credentials(
            self.app_state.authentication_info.user_email,
            self._CLIENT_OAUTH_SECRET,
            self.app_state.authentication_info.id_token,
            self._FIREBASE_DEFAULT_SERVICE_ACCOUNT,
        )

        if not role_credentials:
            st.error(
                "Hubo un error al realizar la autenticación."
                "Intente nuevamente o contacte a soporte."
            )
            log.error(
                "Could not obtain Role Service Account credentials from Firestore"
            )
            return

        self.app_state.authentication_info.role_service_account_credentials = (
            role_credentials
        )

        self.app_state.authentication_info.user_role = role_enum

        log.info(
            f"Successfully authenticated user with email: "
            f"{self.app_state.authentication_info.user_email}"
        )
        st.experimental_rerun()  # Rerun app now authenticated



from abc import ABC, abstractmethod
import pandas as pd
from joblib import load
from dataclasses import dataclass
import tempfile
import os
import json
from google.cloud import storage
from google.oauth2 import service_account
from txp.ml.common.tasks import Policy, PolicyStorage
from txp.common.ml.tasks import AssetTask


@dataclass
class EncoderStorage:
    schema: dict
    dataset: any
    target: str


class Encoder(ABC):
    dataset = None

    def __init__(self, schema):
        self.schema = schema
        self.target = None

    @abstractmethod
    def transform_input(self, signals_completed: dict) -> pd.DataFrame:
        """
            Args:
                signals_completed: filled firestore dictionary, this dict contains all the rows present
                                   in bigquery tables.
            Returns:
                Pandas dataframe which columns schema matches same schema present in the training dataset.
        """
        pass

    @abstractmethod
    def build_training_dataset(self, tables, target):
        pass

    def save(self) -> EncoderStorage:
        return EncoderStorage(self.schema, self.dataset, self.target)

    def load(self, encoder_storage: EncoderStorage):
        print("Restoring encoder")
        self.dataset = encoder_storage.dataset
        self.schema = encoder_storage.schema
        self.target = encoder_storage.target

    def set_schema(self, schema):
        self.schema = schema


@dataclass
class TaskStorage:
    task_id: str
    encoder_storage: EncoderStorage
    policy_storage: PolicyStorage


class Task(ABC):
    task_bucket: str = "txp-model-storage"
    task_id: str

    def __init__(self, encoder: Encoder, policy: Policy, task_definition: AssetTask, credentials_str) -> None:
        self.encoder = encoder
        self.policy = policy
        self.ready = False
        self.credentials = credentials_str
        self.task_definition = task_definition
        self.task_id = self.task_definition.task_id

    def load_from_bucket(self, task_file: str):
        if task_file is not None:
            print(f"Restoring task from {task_file} in {self.task_bucket} bucket")
            tmp_file = download_task_file_from_bucket(self.credentials, self.task_bucket, task_file)
            task_data = load(tmp_file)
            self.load(task_data)
            os.remove(tmp_file)
            self.ready = True

    def save(self) -> TaskStorage:
        return TaskStorage(self.task_id, self.encoder.save(), self.policy.save())

    def load(self, task_storage: TaskStorage):
        self.task_id = task_storage.task_id
        self.encoder.load(task_storage.encoder_storage)
        self.policy.load(task_storage.policy_storage)

    def get_task_id(self):
        return self.task_id

    def get_encoder(self):
        return self.encoder

    def get_schema(self):
        return self.encoder.schema

    def is_ready(self):
        return self.ready

    def set_policy(self, new_policy: Policy):
        self.policy = new_policy

    def transform_dataset(self, dataset) -> None:
        pass

    @abstractmethod
    def predict(self, sampling_window_rows) -> str:
        """Returns the prediction results for the given sampling windows
        generated input vectors.

        Args:
            sampling_window_rows: filled Firestore dictionary, this dict
                contains all the rows present in bigquery tables.

        Returns:
            Enumerated value for Classification Tasks.
        """
        pass


def download_task_file_from_bucket(credentials_str, bucket_name, file_name) -> str:
    json_dict_service_account = json.loads(credentials_str, strict=False)
    credentials = service_account.Credentials.from_service_account_info(json_dict_service_account)
    storage_client = storage.Client(credentials=credentials, project=credentials.project_id)
    bucket = storage_client.bucket(bucket_name)
    blob = bucket.blob(file_name)
    tmp_file = tempfile.NamedTemporaryFile(delete=False)
    blob.download_to_filename(tmp_file.name)

    return tmp_file.name





"""
This module implements the driver host used to manage voyager 3 devices through
the USB smartmesh dongle.
"""
# ============================ imports =========================================
import multiprocessing
from typing import Tuple, List

from txp.devices.drivers.driver import StateReport, DriverState
from txp.devices.drivers.drivers_host import DriversHost
from txp.common.edge import EdgeDescriptor
from txp.devices.package_queue import PackageQueueProxy

from txp.devices.drivers.voyager.smartmesh_dongle import SmartMeshDongle
from txp.devices.drivers.voyager.voyager_driver import VoyagerDriver
from txp.common.edge import DeviceId
from txp.common.config import settings
import logging

log = logging.getLogger(__name__)
log.setLevel(settings.txp.general_log_level)

# ==============================================================================
# VoyagerDriversHost
# ==============================================================================
class VoyagerDriversHost(DriversHost):
    """Concrete implementation of a DriversHost for Voyager edge_descriptors.

    This class is executed as a separated process, and it's responsible for:
        - Start the SmartMeshDongle as smartmesh manager for the motes registered.
        - Host the configured Voyager Drivers for the connected device.
    """

    voyager_driver_errored_timeout_seconds = 240

    def __init__(self, gateway_packages_queue: PackageQueueProxy,drivers_state_queue: multiprocessing.Queue,**kwargs):
        """
        Args:
            gateway_packages_queue: The Gateway package queue to be used by Drivers
                to send collected GatewayPackages.
            **kwargs: named parameters specific to this VoyagerDriversHost implementation.
        """
        super(VoyagerDriversHost, self).__init__(gateway_packages_queue,drivers_state_queue,self.voyager_driver_errored_timeout_seconds,**kwargs)
        log.info("Initialization process of a VoyagerDriversHost")
        self._smartmesh_manager = SmartMeshDongle(self._change_driver_state)

        # Detect the SmartMesh device.
        try:
            log.info("Detecting SmartMesh manager device")
            detected = self._smartmesh_manager.detect()
            if detected:
                log.info("SmartMesh manager device detected.")
            else:
                log.info("SmartMesh manager device not detected.")
        except Exception as e:
            log.error(f"Exception detecting SmartMesh device: {e}")

    def connect(self):
        """Connects the Voyager 3 Dongle unit"""
        try:
            self._smartmesh_manager.connect()
        except Exception as e:
            log.error(f"Exception connecting smartmesh devices: {e}")

        if not self._smartmesh_manager.connected:
            raise Exception("Host failed to initiailze connection with the device.")

    def disconnect(self):
        """Disconnects the Voyager 3 dongle unit"""
        self._smartmesh_manager.disconnect()

    def discover_physical_ids(self) -> List[DeviceId]:
        """Returns the set of found DeviceId in the smartmesh network.

        Returns:
            The List of DeviceId.
        """
        result = list()

        for mac in self._smartmesh_manager.operational_macs:
            dev_id = DeviceId("", SmartMeshDongle.pretty_mac(mac))
            result.append(dev_id)

        print(result)

        return result

    def get_driver_host_name(self):
        return VoyagerDriver.device_name()

    def _add_drivers(self, edges: List[EdgeDescriptor]):
        """Adds the set of drivers for the specified edges in the host.

        Args:
            edges: List of EdgeDescriptor's for the drivers to be created in
                the host instance.
        """
        for descriptor in edges:
            driver = VoyagerDriver(descriptor, self.gateway_package_queue, self._smartmesh_manager,
                                   self._on_driver_state_changed, self._completed_driver_pulses_queue)
            self.drivers.append(driver)

    #TODO: this has to be changed. This method will be putted in base class and physical_id in this case is named mac will be standardized for both drivers since thi code
    # is being repeated in both hosts. same internal content.
    def _change_driver_state(self, mac: Tuple, driver_state):
        drivers = list(filter(lambda driver : driver.mac == mac,self.drivers))
        if len(drivers) > 0:
            driver = drivers.pop()
            if driver_state == DriverState.DISCONNECTED:
                driver.stop_sampling_windows()
            driver.update_driver_state(driver_state)
            self._add_drivers_state_to_queue(StateReport(driver.logical_id, driver_state))
        else:
            log.error("{} DEVICE IS NOT IN THE CURRENT  GATEWAY CONFIGURATION. ILLEGAL CONNECTION FROM SERIAL {}".format(self.get_driver_host_name(),mac))

    def take_fix_error_action(self, serial):
        pass




"""
This module implements an enumerated registry of known devices and their
driver implementation.
"""
from enum import Enum

from txp.devices.drivers.arm_robot.thermal_robot import ThermalRobotDriver
from txp.devices.drivers.icomox.icomox_driver import IcomoxDriver
from txp.devices.drivers.icomox.icomox_driver_host import IcomoxDriversHost
from txp.devices.drivers.mock.mock_driver import MockDriver, MockDriversHost
from txp.devices.drivers.voyager.voyager_driver import VoyagerDriver
from txp.devices.drivers.voyager.voyager_driver_host import VoyagerDriversHost
from txp.devices.drivers.usb.usb_base_driver import UsbDriver
from txp.devices.drivers.usb.usb_driver_host import UsbDriversHost
from txp.devices.drivers.thermocouple.mcp9600_driver import Mcp9600Driver
from txp.devices.drivers.generic_driver_host import GenericDriversHost
from txp.devices.drivers.arm_robot.robot_driver import ArmRobotDriver
from txp.devices.drivers.arm_robot.robot_host import RobotDriversHost, ThermalArmHost
from txp.devices.drivers.driver import Driver
from txp.devices.drivers.drivers_host import DriversHost
from txp.devices.package_queue import PackageQueueProxy
from multiprocessing.managers import BaseManager
from txp.common.config import settings
import logging

log = logging.getLogger(__name__)
log.setLevel(settings.txp.general_log_level)


# ==============================================================================
# Enumeration for the known communication device drivers.
# ==============================================================================
class DeviceType(Enum):
    """Enumeration registry for the different types of defined driver implementations.
    The _value_ for the enumeration is the name of the device defined by each driver
    with the class method device_name().

    The registry contains for each value:
        - The Driver class for the device.
        - The DriversHost class for the device.
        - Defined methods that helps with different use cases.

    """

    def __new__(cls, driver_class: Driver, devices_host_class: DriversHost):
        obj = object.__new__(cls)
        obj._value_ = driver_class.device_name()
        obj.driver = driver_class
        obj.host = devices_host_class

        return obj

    def build_host_process(self, process_manager: BaseManager, package_queue: PackageQueueProxy, drivers_state_queue,
                           **kwargs) -> DriversHost:
        """Builds a server process DriverHost instance using the provided
        process_manager.

        The others parameters after the process_manager, are the parameters expected by the
        constructor of a DriversHost class.

        Args:
            process_manager: The custom multiprocessing.managers.BaseManager that contains
                the DriversHost classes registered. The process_manager should had been started using
                the method start().
            package_queue: The Gateway's PackageQueue expected by each Drivers Host.
            **kwargs: The named kwargs expected by the DriversHost __init__ method.
            :param drivers_state_queue:
        """
        host_constructor = getattr(process_manager, self.host.__name__)
        try:
            host_instance_proxy = host_constructor(package_queue, drivers_state_queue, **kwargs)
        except Exception as ex:
            log.error(ex)

        return host_instance_proxy

    VOYAGER = (VoyagerDriver, VoyagerDriversHost)
    ICOMOX = (IcomoxDriver,IcomoxDriversHost)
    MOCK = (MockDriver, MockDriversHost)
    USB = (UsbDriver, UsbDriversHost)
    THERMOCOUPLE = (Mcp9600Driver, GenericDriversHost)
    ARMROBOT = (ArmRobotDriver, RobotDriversHost)
    THERMALROBOT = (ThermalRobotDriver, ThermalArmHost)



"""
This module contains classes related to the definition and usage of the Sampling
Configuration defined for a gateway device.
"""

# ============================ imports =========================================

from txp.common.configuration import SamplingWindow
from txp.common.configuration import SamplingParameters


class SamplingTask:
    """A sampling task to be performed by a Driver.

    This structure works as an abstraction to contain any kind of information required
    by the driver to perform sampling tasks.
    """

    def __init__(self, configuration_id: str, parameters: SamplingParameters, tenant_id: str):
        """
        Args:
            configuration_id: the configuration id under which this task was assigned.
            task_configuration: The SamplingParameters for this task.
            tenant_id (str): The tenant_id to which this Gateway belongs
        """
        self.sampling_window: SamplingWindow = parameters.sampling_window
        self.parameters: SamplingParameters = parameters
        self.configuration_id: str = configuration_id
        self.tenant_id: str = tenant_id



"""
This module defines the main function to be called from the CLI program
for the command: run_voyager_test
"""
# ============================ imports =========================================
from txp.devices.drivers.voyager.voyager_driver import VoyagerDriver
from txp.devices.drivers.voyager.voyager_driver_host import VoyagerDriversHost
from txp.common.edge import EdgeDescriptor, EdgeType, MachineMetadata
from txp.devices.sampling_task import SamplingTask
from txp.common.configuration import SamplingWindow
from txp.devices.package_queue import PackageQueue, PackageQueueProxy
import multiprocessing
from typing import Tuple
import sys
import time
from txp.common.config import settings
import logging

log = logging.getLogger(__name__)
log.setLevel(settings.txp.general_log_level)


# ==============================================================================
# Server process to MOCK the gateway's package queue.
# ==============================================================================
class ProcessesManager(multiprocessing.managers.BaseManager):
    """This class will be used to create the Driver, the SmartMeshManager and the
    PackageQueue in separated processes"""

    pass


ProcessesManager.register("PackageQueue", PackageQueue, proxytype=PackageQueueProxy)

# ==============================================================================
# Declares the EdgeDescriptors for the Voyager instance
# ==============================================================================
machine_metadata: MachineMetadata = MachineMetadata(
    "licuadora_1", "Clasica 4655", "Osterizer", ["voyager_1"]
)

voyager_1_descriptor: EdgeDescriptor = EdgeDescriptor(
    "voyager_1",
    EdgeType.STREAM_ONLY_DEVICE,
    VoyagerDriver.device_name(),
    {},
    {
        "VibrationAcceleration": {
            VoyagerDriver.PERCEPTION_DIMENSIONS_FIELD: [3, 512],
            VoyagerDriver.PERCEPTION_SAMPLING_FREQUENCY_FIELD: 5000
        }
    }
)

# ==============================================================================
# Declares the SamplingWindow for the Voyager instance
# Observation time: 40s
# Sampling time: 20s
# ==============================================================================
sampling_window: SamplingWindow = SamplingWindow(40, 35)


def voyager_test(mac: Tuple):
    package_queue_proc_manager = ProcessesManager()

    package_queue_proc_manager.start()

    gateway_packages_queue: PackageQueueProxy = package_queue_proc_manager.PackageQueue()

    voyagers_host = VoyagerDriversHost(gateway_packages_queue)

    voyagers_host.connect()

    # The driver expects the MAC address as a parameter.
    # This MAC address setup is a responsibility of the Gateway
    # when it's processing the operations configuration.

    voyager_1_descriptor.edge_parameters["mac"] = mac

    voyagers_host.add_drivers([voyager_1_descriptor])

    voyagers_host.connect_drivers()
    devicesId = voyagers_host.discover_physical_ids()

    # voyager_1_descriptor 1 sampling task assigned to collect 2 packages.
    task1 = SamplingTask(2, sampling_window, 'client')
    voyagers_host.receive_sampling_tasks(
        {
            voyager_1_descriptor.logical_id: task1
        }
    )

    voyagers_host.start_sampling_windows()
    while True:
        time.sleep(10)
        #voyagers_host.stop_sampling_windows()

if __name__ == "__main__":
    """Main function to debug if the module is executed."""
    args = sys.argv[1:]

    mac_string = args[0].split("-")
    mac = tuple(int(i,16) for i in mac_string)

    voyager_test(mac)



"""XML utility functions"""

from xml.dom.minidom import parseString


def xml_obj_to_dict(xml_obj):
    attrs = {}
    ch_node = xml_obj.firstChild
    while ch_node:
        # look for ELEMENT_NODEs
        if ch_node.nodeType == ch_node.ELEMENT_NODE:
            #print 'Reading', ch_node.nodeName
            n = str(ch_node.nodeName)
            v = ''
            # iterate through the children of the current node either
            # accumulating multiple child text nodes into a single string, or
            # recursing to build a dict of child elements
            ch2_node = ch_node.firstChild
            while ch2_node:
                if ch2_node.nodeType in [ch_node.TEXT_NODE, ch_node.CDATA_SECTION_NODE]:
                    v += str(ch2_node.nodeValue)

                # if the child is an element, then build a dict
                elif ch2_node.nodeType == ch_node.ELEMENT_NODE:
                    v = xml_obj_to_dict(ch_node)
                    break

                ch2_node = ch2_node.nextSibling

            # populate the attributes dict with the value
            # create a list of values if the child name already exists
            if n in attrs:
                if not type(attrs[n]) is list:
                    attrs[n] = [attrs[n]]
                attrs[n].append(v)
            else:
                attrs[n] = v
            if ch_node.hasAttributes():
                for name, val in ch_node.attributes.items():
                    attr_name = n + name.capitalize()
                    attrs[attr_name] = val

        ch_node = ch_node.nextSibling
    
    return attrs


def parse_xml_obj(xml_str, base_element, fields = None):
    xml_doc = parseString(xml_str)
    obj_docs = xml_doc.getElementsByTagName(base_element)
    # parse each found object into a dict of fields
    obj_dicts = []
    for obj_doc in obj_docs:
        obj_dicts.append(xml_obj_to_dict(obj_doc))
        if fields:
            # TODO: filter dict for requested fields
            pass
    return obj_dicts

def _dict_to_xml(inpDict, outList) :
    for k in inpDict :
        outList.append('<{0}>'.format(k))
        if inpDict[k] :
            if isinstance(inpDict[k], dict) :
                _dict_to_xml(inpDict[k], outList)
            else :
                outList.append('{0}'.format(inpDict[k]))
        outList.append('</{0}>'.format(k))

def dict_to_xml(inpDict, prefix = None) :
    outList = []
    if prefix :
        outList.append(''.join(['<{0}>'.format(p) for p in prefix]))
    _dict_to_xml(inpDict, outList)
    if prefix :
        outList.append(''.join(['</{0}>'.format(p) for p in reversed(prefix)]))
    return ''.join([l for l in outList])

def list_to_xml(inpList, el_tag, prefix = None):
    outList = []
    if prefix :
        outList.append(''.join(['<{0}>'.format(p) for p in prefix]))
    for el in inpList:
        outList.append('<{0}>{1}</{0}>'.format(el_tag, el))
    if prefix :
        outList.append(''.join(['</{0}>'.format(p) for p in reversed(prefix)]))
    return ''.join(outList)



import apache_beam as beam
import logging


class TimeProcessing(beam.DoFn):

    def to_runner_api_parameter(self, unused_context):
        return "beam:transforms:custom_parsing:custom_v0", None

    def process(self, element, timestamp=beam.DoFn.TimestampParam, window=beam.DoFn.WindowParam):
        import google.cloud.firestore as firestore
        from txp.cloud.common import utils as cu

        firestore_db = firestore.Client()
        mode = cu.get_signal_mode_from_firestore(element["configuration_id"], element["tenant_id"],
                                                 element["edge_logical_id"], element["perception_name"], firestore_db)
        if mode is None or not cu.SignalMode.is_time(mode):
            return

        yield element


class FftProcessing(beam.DoFn):

    def to_runner_api_parameter(self, unused_context):
        return "beam:transforms:custom_parsing:custom_v0", None

    def process(self, element, timestamp=beam.DoFn.TimestampParam, window=beam.DoFn.WindowParam):
        from scipy.fft import fft
        import google.cloud.firestore as firestore
        from txp.cloud.common import utils as cu

        firestore_db = firestore.Client()
        mode = cu.get_signal_mode_from_firestore(element["configuration_id"], element["tenant_id"],
                                                 element["edge_logical_id"], element["perception_name"], firestore_db)
        if mode is None or not cu.SignalMode.is_fft(mode):
            return
        data = [list(dimension["values"]) for dimension in element["data"]]
        fft_data = []
        for index, dimension_signal_sample in enumerate(data):
            fft_data.append({
                "values": [],
                "index": index
            })
            n = len(fft_data)
            for z in fft(dimension_signal_sample):
                fft_data[n - 1]["values"].append({"real": float(z.real), "imag": float(z.imag)})

        yield {
            "fft": fft_data,
            "package_timestamp": element["package_timestamp"],
            "perception_name": element["perception_name"],
            "edge_logical_id": element["edge_logical_id"],
            "signal_timestamp": element["signal_timestamp"],
            "configuration_id": element["configuration_id"],
            "observation_timestamp": element["observation_timestamp"],
            "gateway_task_id": element["gateway_task_id"],
            "sampling_window_index": element["sampling_window_index"],
            "number_of_sampling_windows": element["number_of_sampling_windows"],
            "tenant_id": element["tenant_id"],
            "partition_timestamp": element["partition_timestamp"]
        }


class PsdProcessing(beam.DoFn):

    def to_runner_api_parameter(self, unused_context):
        return "beam:transforms:custom_parsing:custom_v0", None

    def process(self, element, timestamp=beam.DoFn.TimestampParam, window=beam.DoFn.WindowParam):
        import google.cloud.firestore as firestore
        from txp.cloud.common import utils as cu
        from txp.cloud.common import metrics

        firestore_db = firestore.Client()
        mode = cu.get_signal_mode_from_firestore(element["configuration_id"], element["tenant_id"],
                                                 element["edge_logical_id"], element["perception_name"], firestore_db)
        if mode is None or not cu.SignalMode.is_psd(mode):
            return
        data = [list(dimension["values"]) for dimension in element["data"]]
        data_psd = []
        for index, dimension_signal_sample in enumerate(data):
            f, psd = metrics.get_psd(dimension_signal_sample, cu.get_sampling_frequency(element, firestore_db))
            data_psd.append({
                "psd": [float(e) for e in psd],
                "frequency": [float(e) for e in f],
                "index": index,
            })

        yield {
            "data": data_psd,
            "package_timestamp": element["package_timestamp"],
            "perception_name": element["perception_name"],
            "edge_logical_id": element["edge_logical_id"],
            "signal_timestamp": element["signal_timestamp"],
            "configuration_id": element["configuration_id"],
            "observation_timestamp": element["observation_timestamp"],
            "gateway_task_id": element["gateway_task_id"],
            "sampling_window_index": element["sampling_window_index"],
            "number_of_sampling_windows": element["number_of_sampling_windows"],
            "tenant_id": element["tenant_id"],
            "partition_timestamp": element["partition_timestamp"]
        }


class TimeMetrics(beam.DoFn):

    def to_runner_api_parameter(self, unused_context):
        return "beam:transforms:custom_parsing:custom_v0", None

    def process(self, element, timestamp=beam.DoFn.TimestampParam, window=beam.DoFn.WindowParam):
        from txp.cloud.common import metrics
        import google.cloud.firestore as firestore
        from txp.cloud.common import utils as cu

        firestore_db = firestore.Client()
        mode = cu.get_signal_mode_from_firestore(element["configuration_id"], element["tenant_id"],
                                                 element["edge_logical_id"], element["perception_name"], firestore_db)
        if mode is None or cu.SignalMode.is_image(mode):
            return

        for dimension in range(0, len(element["data"])):
            yield {
                "observation_timestamp": element["observation_timestamp"],
                "edge_logical_id": element["edge_logical_id"],
                "perception_name": element["perception_name"],
                "signal_timestamp": element["signal_timestamp"],
                "package_timestamp": element["package_timestamp"],
                "dimension": dimension,
                "configuration_id": element["configuration_id"],
                "peak": float(metrics.peak(element["data"][dimension]["values"])),
                "rms": float(metrics.rms(element["data"][dimension]["values"])),
                "standard_deviation": float(metrics.standard_deviation(element["data"][dimension]["values"])),
                "crest_factor": float(metrics.crest_factor(element["data"][dimension]["values"])),
                "gateway_task_id": element["gateway_task_id"],
                "sampling_window_index": element["sampling_window_index"],
                "number_of_sampling_windows": element["number_of_sampling_windows"],
                "tenant_id": element["tenant_id"],
                "partition_timestamp": element["partition_timestamp"]
            }


class FftMetrics(beam.DoFn):

    def to_runner_api_parameter(self, unused_context):
        return "beam:transforms:custom_parsing:custom_v0", None

    def process(self, element, timestamp=beam.DoFn.TimestampParam, window=beam.DoFn.WindowParam):
        from txp.cloud.common import metrics
        from txp.cloud.common import utils as cu

        data_per_dimension = cu.get_fft_as_np_array(element["fft"])
        for dimension in range(0, len(data_per_dimension)):
            data = data_per_dimension[dimension].real
            yield {
                "edge_logical_id": element["edge_logical_id"],
                "perception_name": element["perception_name"],
                "signal_timestamp": element["signal_timestamp"],
                "package_timestamp": element["package_timestamp"],
                "dimension": dimension,
                "configuration_id": element["configuration_id"],
                "peak": float(metrics.peak(data)),
                "rms": float(metrics.rms(data)),
                "standard_deviation": float(metrics.standard_deviation(data)),
                "crest_factor": float(metrics.crest_factor(data)),
                "observation_timestamp": element["observation_timestamp"],
                "gateway_task_id": element["gateway_task_id"],
                "sampling_window_index": element["sampling_window_index"],
                "number_of_sampling_windows": element["number_of_sampling_windows"],
                "tenant_id": element["tenant_id"],
                "partition_timestamp": element["partition_timestamp"]
            }


class PsdMetrics(beam.DoFn):

    def to_runner_api_parameter(self, unused_context):
        return "beam:transforms:custom_parsing:custom_v0", None

    def process(self, element, timestamp=beam.DoFn.TimestampParam, window=beam.DoFn.WindowParam):
        from txp.cloud.common import metrics
        import numpy as np

        for dimension in range(0, len(element["data"])):
            f = element["data"][dimension]["frequency"]
            psd = element["data"][dimension]["psd"]
            integrated_psd = metrics.integrate(psd, x=f)
            yield {
                "observation_timestamp": element["observation_timestamp"],
                "edge_logical_id": element["edge_logical_id"],
                "perception_name": element["perception_name"],
                "signal_timestamp": element["signal_timestamp"],
                "package_timestamp": element["package_timestamp"],
                "dimension": dimension,
                "configuration_id": element["configuration_id"],
                "rms": float(np.sqrt(integrated_psd[-1])),
                "standard_deviation": float(metrics.standard_deviation(psd)),
                "crest_factor": float(metrics.crest_factor(psd)),
                "peak_frequency": float(metrics.peak(f)),
                "peak_amplitude": float(metrics.peak(psd)),
                "gateway_task_id": element["gateway_task_id"],
                "sampling_window_index": element["sampling_window_index"],
                "number_of_sampling_windows": element["number_of_sampling_windows"],
                "tenant_id": element["tenant_id"],
                "partition_timestamp": element["partition_timestamp"]
            }


class WriteToBigQuery(beam.DoFn):

    def to_runner_api_parameter(self, unused_context):
        return "beam:transforms:custom_parsing:custom_v0", None

    def process(self, element, table_id, timestamp=beam.DoFn.TimestampParam, window=beam.DoFn.WindowParam):
        from google.cloud import bigquery
        from datetime import datetime
        import json

        keys_to_ignore = ["gateway_task_id", "sampling_window_index", "number_of_sampling_windows"]
        bigquery_element = {k: element[k] for k in element if k not in keys_to_ignore}
        client = bigquery.Client()
        table_id = table_id.replace(":", ".")
        errors = client.insert_rows_json(table_id, [bigquery_element])
        message = f"""Edge: {element["edge_logical_id"]}
        Perception: {element["perception_name"]}
        Signal Timestamp: {datetime.fromtimestamp(
            element["signal_timestamp"] // 1e9).strftime('%Y-%m-%d %H:%M:%S')}
        Package Timestamp: {datetime.fromtimestamp(
            element["package_timestamp"] // 1e9).strftime('%Y-%m-%d %H:%M:%S')}
        """
        if not errors:
            logging.info(f"""Storing in {table_id}: {message} """)

            res = {
                "table_id": table_id,
                "edge_logical_id": element["edge_logical_id"],
                "perception_name": element["perception_name"],
                "package_timestamp": element["package_timestamp"],
                "signal_timestamp": element["signal_timestamp"],
                "observation_timestamp": element["observation_timestamp"],
                "configuration_id": element["configuration_id"],
                "gateway_task_id": element["gateway_task_id"],
                "sampling_window_index": element["sampling_window_index"],
                "number_of_sampling_windows": element["number_of_sampling_windows"],
                "tenant_id": element["tenant_id"],
                "partition_timestamp": element["partition_timestamp"]
            }
            if "previous_part_index" in element:
                res["previous_part_index"] = element["previous_part_index"]
            if "part_index" in element:
                res["part_index"] = element["part_index"]
            if "dimension" in element:
                res["dimension"] = element["dimension"]

            res = json.dumps(res).encode('utf-8')

            yield res
        else:
            logging.error(f"""Could not store in {table_id}: {message}, ERROR: {errors} """)



"""
This module defines the main function to be called from the CLI program
for the command: ru_device_manager
"""
# ============================ imports =========================================

from txp.devices.devices_manager import DevicesManager
from txp.devices.drivers.icomox.icomox_driver import IcomoxDriver
from txp.common.edge import EdgeDescriptor, EdgeType, MachineMetadata
from txp.devices.sampling_task import SamplingTask
from txp.common.configuration import SamplingWindow
from txp.devices.package_queue import PackageQueue, PackageQueueProxy
from txp.common.configuration import SamplingParameters
import multiprocessing
from typing import Tuple
import datetime
import sys
import time

from txp.common.config import settings
import logging

log = logging.getLogger(__name__)
log.setLevel(settings.txp.general_log_level)


# ==============================================================================
# Server process to MOCK the gateway's package queue.
# ==============================================================================
class ProcessesManager(multiprocessing.managers.BaseManager):
    """This class will be used to create the Driver, the SmartMeshManager and the
    PackageQueue in separated processes"""

    pass


ProcessesManager.register("PackageQueue", PackageQueue, proxytype=PackageQueueProxy)

# ==============================================================================
# Declares the EdgeDescriptors for the Icomox instance
# ==============================================================================
machine_metadata: MachineMetadata = MachineMetadata(
    "licuadora_1", "Clasica 4655", "Osterizer", ["icomox_1"]
)

icomox_1_descriptor: EdgeDescriptor = EdgeDescriptor(
    "icomox_1",
    EdgeType.STREAM_ONLY_DEVICE,
    IcomoxDriver.device_name(),
    {
        IcomoxDriver._PHYSICAL_ID_PARAM_FIELD : "12A79EA89FFE7C394EFC030038343038",
    },
    {
        "VibrationAcceleration": {
            IcomoxDriver.PERCEPTION_DIMENSIONS_FIELD: [3, 1024],
            IcomoxDriver.PERCEPTION_SAMPLING_FREQUENCY_FIELD: 3611.11
        },
        "VibrationSpeed": {
            IcomoxDriver.PERCEPTION_DIMENSIONS_FIELD: [3, 512],
            IcomoxDriver.PERCEPTION_SAMPLING_FREQUENCY_FIELD: 399.60
        },
        "Magnetometer": {
            IcomoxDriver.PERCEPTION_DIMENSIONS_FIELD: [3, 512],
            IcomoxDriver.PERCEPTION_SAMPLING_FREQUENCY_FIELD: 334.36
        },
        "Microphone": {
            IcomoxDriver.PERCEPTION_DIMENSIONS_FIELD: [1, 512],
            IcomoxDriver.PERCEPTION_SAMPLING_FREQUENCY_FIELD: 20312.5
        }
    }
)


# ==============================================================================
# Declares the SamplingWindow
# Observation time: 40s
# Sampling time: 20s
# ==============================================================================
sampling_window: SamplingWindow = SamplingWindow(40, 35)
start_date = datetime.datetime.now().date()
end_date = start_date + datetime.timedelta(days=1)
start_time = (datetime.datetime.now() + datetime.timedelta(seconds=15)).time()
end_time = (datetime.datetime.now() + datetime.timedelta(minutes=4)).time()
sampling_parameters: SamplingParameters = SamplingParameters(
    sampling_window,
    start_date,
    end_date,
    "1111111",
    start_time,
    end_time,
    []
)
state_queue = multiprocessing.Manager().Queue()
tasks = {
    icomox_1_descriptor.logical_id:SamplingTask("0", sampling_parameters, "labshowroom-001")
}


def device_manager_test(unique_id: Tuple):
    devices_manager = DevicesManager()
    try:
        package_queue_proc_manager = ProcessesManager()
        package_queue_proc_manager.start()
        gateway_packages_queue: PackageQueueProxy = package_queue_proc_manager.PackageQueue()

        #device manager initialization
        devices_manager.create_host(IcomoxDriver.device_name(),gateway_packages_queue)
        devices_manager.initialize_drivers([icomox_1_descriptor])
        devices_manager.start()
        devices_manager.receive_sampling_tasks(tasks)

        #for this example I will only wait for the first device to report it is connected in order to start sampling
        #this method should be used in the backgound to update the states of drivers table handled by the gateway.
        while True:
            states = devices_manager.get_state_changed_queued_items()
            if len(states) > 0:
                break

        devices_manager.start_sampling_windows()

        while True:
            time.sleep(10)
    finally:
        devices_manager.stop()

if __name__ == "__main__":
    """Main function to debug if the module is executed."""
    args = sys.argv[1:]

    unique_id = tuple(args[0].split("-"))
    device_manager_test(unique_id)



"""
This module implements an Integrated Analytics View for the application.

NOTE: This is just a dummy implementation of a MainView in order to shown the
mechanism provided by the `core_components` package.
"""

import pytz
from txp.cloud.controllers.authentication import AuthenticationInfo
from txp.web.core_components.app_profiles import AppProfile
from txp.web.core_components.main_view import MainView
from typing import Dict
import streamlit as st
from txp.web.core_components.annotation_widget import AnnotationWidget
import json
from txp.common.config import settings
import txp.cloud.controllers.bigquery as bigquery
from pylab import *
import txp.web.views.analytics.computer_vision_analytics as computer_vision_utils
import txp.web.views.analytics.vibration_analytics as vibration_utils
import txp.web.views.analytics.complex_analytics as complex_utils
import txp.web.views.analytics.temperature_analytics_view as temperature_utils
import plotly.graph_objects as go
from dataclasses import dataclass, field
from google.cloud.bigquery import Client
import pandas as pd

log = logging.getLogger(__name__)
log.setLevel(settings.txp.general_log_level)

_MAX_IMAGES_ALLOWED_RANGE = 3600
_IMAGES_RANGE_TIME_EXCEEDED = False


@dataclass
class DataRequest:
    """
    DataRequest manages the different data requests needed for render content in this view.
    Args:
         range_time: date, start and end time for data request
         edge_kind(str): edge kind from selected edge for data request
         tz: timezone
         bigquery_dataset(str): bigquery dataset name
         tenant_id(str): tenant id name
         edge(str): edge name for data request
         perception(str): perception name for data request
         bqclient: bigquery client for performing queries
         order_key(str): order key used in bigquery.get_all_records_within_interval() method


    """

    range_time: Dict[str, datetime.datetime] = field(default_factory=dict)
    edge_kind: str = None
    tz = pytz.timezone("America/Mexico_City")
    bigquery_dataset: str = None
    tenant_id: str = None
    edge: str = None
    perception: str = None
    bqclient: Client = None
    order_key: str = "observation_timestamp"

    def _get_full_start_date(self):
        """
        returns the full start date localized
        """
        full_start_date = self.tz.localize(
            datetime.datetime.combine(self.range_time["date"], self.range_time["start"])
        )
        return full_start_date.astimezone(pytz.utc)

    def _get_full_end_date(self):
        """
        returns the full end date localized
        """
        full_end_date = self.tz.localize(
            datetime.datetime.combine(self.range_time["date"], self.range_time["end"])
        )
        return full_end_date.astimezone(pytz.utc)

    def get_data_from_table_in_dataset(self, table, user_role: AuthenticationInfo.UserRole):
        """
        returns the data from a table if data, else return None
        """
        full_start_date = self._get_full_start_date()
        full_end_date = self._get_full_end_date()
        global _IMAGES_RANGE_TIME_EXCEEDED
        # Don't allow Client users to download pictures from 1 hour or more.
        if (
            user_role != AuthenticationInfo.UserRole.Admin
            and (full_end_date - full_start_date).total_seconds() > _MAX_IMAGES_ALLOWED_RANGE
            and self.perception in { 'Image', 'ThermalImage' }
        ):
            log.info("Client user trying to get Images over an hour")
            _IMAGES_RANGE_TIME_EXCEEDED = True
            return None
        elif self.perception in { 'Image', 'ThermalImage' }:
            _IMAGES_RANGE_TIME_EXCEEDED = False

        signals_from_interval = self._get_data_from_table_in_dataset(table)
        if not signals_from_interval or len(signals_from_interval) == 0:
            logging.info(f"No hay data disponible")
            return None
        else:
            return signals_from_interval

    def _get_data_from_table_in_dataset(self, table):
        """
        returns the data in a table from a bigquery dataset.
        Args:
            table(str): bigquery table name.
        """
        full_start_date = self._get_full_start_date()
        full_end_date = self._get_full_end_date()
        return bigquery.get_all_signals_within_interval(
            tenant_id=self.tenant_id,
            table_name=f"{self.bigquery_dataset}.{table}",
            edge_logical_id=self.edge,
            perception_name=self.perception,
            start_datetime=full_start_date,
            end_datetime=full_end_date,
            client=self.bqclient,
        )

    def _get_data_from_table_in_dataset_ordered_by_value(
            self, table
    ):
        """
        returns the data in a table from a bigquery dataset ordered by the key included as self.order_by.
        Args:
            table(str): bigquery table name.
        """
        full_start_date = self._get_full_start_date()
        full_end_date = self._get_full_end_date()

        df = bigquery.get_all_records_within_interval(
            tenant_id=self.tenant_id,
            table_name=f"{self.bigquery_dataset}.{table}",
            edge_logical_id=self.edge,
            perception_name=self.perception,
            start_datetime=full_start_date,
            end_datetime=full_end_date,
            order_key=self.order_key,
            client=self.bqclient,
        )
        if df.empty:
            return None
        else:
            return df


@dataclass
class VibrationDataCharts:
    fft_chart: go.Figure = None
    fft_metrics_df: pd.DataFrame = None
    psd_chart: go.Figure = None
    psd_metrics_df: pd.DataFrame = None


@dataclass
class ImagesDataCharts:
    normal_imgs_chat: go.Figure = None
    thermographic_imgs_datacharts: go.Figure = None
    chart_length_exceeded: bool = False


@dataclass
class TemperatureDataCharts:
    temp_chart: go.Figure = None


class AnalyticsView(MainView):
    """The Analytics View"""

    def __init__(self):
        super(AnalyticsView, self).__init__(component_key="analytics_view")
        self.widget = None
        self.current_machine = None
        self.current_machines_groups = None
        self.data = None
        self.data_request: DataRequest
        self.edges = None
        self.edge = None
        self.current_readable_names_map = None
        self.formatted_asset_name = None
        self.edge_kind = None
        self.current_perception = None
        self.sheet = None
        # attributes related to images
        self.data_format = None
        self._animation_data_format = "Animación"
        self._images_data_format = "Imágenes"
        self._timages_data_format = "Imágenes termográficas"
        self._thermographic_perception = "ThermalImage"
        self.normal_data = None
        self.thermographic_data = None
        self._image_perception = "Image"
        self._images_kind = {"ArmRobot"}
        # attributes related to vibration
        self.edges_data = None
        self.vibrational_perceptions = None
        self.vibration_data = None
        self.fft_metrics_df = None
        self.psd_metrics_df = None
        self._vibrational_kind = {"Icomox"}
        # attribuets related to complex and temperature
        self._complex_kind = {"ThermoArm"}
        self._temperature_kind = {"Generic"}
        self._temperature_perception = "Temperature"
        self.temperature_data = None
        # attribuets related to temperature
        self.last_temperatures_values = None
        self._per_time_sheet = "Registro por tiempo"
        self._per_last_signals = "Últimos registros"
        self._per_condition = "Registro por condición"

        # The control attributes defined below helps in performance because allows to
        #   know when is necessary to repaint something.
        self._redraw_required: bool = False
        self._vibration_charts: VibrationDataCharts = VibrationDataCharts()
        self._images_charts: ImagesDataCharts = ImagesDataCharts()
        self._temperature_charts: TemperatureDataCharts = TemperatureDataCharts()

    def get_associated_profile(self) -> AppProfile:
        return AppProfile.ANALYTICS

    @classmethod
    def get_display_name(cls) -> str:
        return "Análisis de Datos"

    def _build_submenu_state(self) -> Dict:
        pass

    def _build_content_state(self) -> Dict:
        pass

    def get_kind_from_edge(self):
        kind = self.data["edges"][self.edge]["device_kind"]
        return kind

    def formatted_data(self, data):
        """
        takes the project data and returns a formatted Dict.
        """
        machines = {d.machine_id: d.reprJSON() for d in data.machines_table.values()}
        gateways = {d.gateway_id: d.reprJSON() for d in data.gateways_table.values()}
        edges = {d.logical_id: d.reprJSON() for d in data.edges_table.values()}
        machines_groups = {d["name"]: d for d in data.machines_groups_table.values()}
        f_data = {
            "machines_groups": machines_groups,
            "machines": machines,
            "gateways": gateways,
            "edges": edges,
        }
        return f_data

    def names_human_readable(self):
        """
        filters the human readable map with the current machines edges
        """
        readable_names_map = {}
        readable_names_map_list = json.loads(st.secrets["edges_names"])
        for k, v in readable_names_map_list.items():
            if k in self.edges:
                readable_names_map.update({k: v})
        return readable_names_map

    def get_data(self):
        if self.data is None:
            self.data = self.formatted_data(self.app_state.project_model)
            machines = list(self.data["machines"].keys())
            machines_groups = list(self.data["machines_groups"].keys())
            self.current_machines_groups = machines_groups[0]
            self.current_machine = machines[0]
            self.edges = self.get_edges_from_asset(
                self.data["machines"][self.current_machine]
            )
            self.current_readable_names_map = self.names_human_readable()
            if not self.edges:
                self.edge = None
            else:
                self.edge = self.edges[0]
                self.edge_kind = self.get_kind_from_edge()
            self.sheet = self._per_time_sheet
            self.bqclient = bigquery.get_authenticated_client(
                self.app_state.authentication_info.role_service_account_credentials
            )
            self.edges_data = {
                machine.machine_id: vibration_utils.mapping_data_from_data_parent(
                    self.app_state.project_model.edges_table.values()
                )
                for machine in self.app_state.project_model.machines_table.values()
            }

            self.data_request = DataRequest()

            if self.edge_kind in self._vibrational_kind:
                self.vibrational_analysis_assets = (
                    vibration_utils.get_machines_with_vibration_analysis(
                        self.edges_data
                    )
                )
                self.vibrational_perceptions = vibration_utils.vibrational_perceptions(
                    edges_data=self.edges_data,
                    current_machine=self.current_machine,
                    current_edge=self.edge,
                    project_data=self.app_state.project_model,
                )
                self.current_perception = self.vibrational_perceptions[0]
            elif self.edge_kind in self._images_kind:
                self.data_format = self._animation_data_format
            elif self.edge_kind in self._complex_kind:
                self.data_format = self._animation_data_format
            else:
                log.info("No data updated")

    def get_edges_from_asset(self, current_asset):
        edges = []
        for edge_name in current_asset["edges_ids"]:
            edge_type = (
                self.data['edges'].get('type', None)
            )  # there are edges with device_type and type in firestore
            if edge_type:
                if edge_type == "ACTUATOR_DEVICE":
                    pass
                else:
                    edges.append(edge_name)
            else:
                edge_type = (
                    self.data['edges'].get('device_type', None)
                )  # there are edges with device_type and type in firestore
                if edge_type == "ACTUATOR_DEVICE":
                    pass
                else:
                    edges.append(edge_name)
        return edges

    def get_perceptions_for_robot_devices(self):
        """
        In robot devices, returns a list with the perceptions according to the current edge kind.
        """
        if self.edge_kind in self._images_kind:
            return [self._image_perception, self._thermographic_perception]
        elif self.edge_kind in self._complex_kind:
            return [
                self._image_perception,
                self._thermographic_perception,
                self._temperature_perception,
            ]

    def _render_submenu(self):
        self.get_data()
        machines = list(self.data["machines"].keys())
        timezone = pytz.timezone("America/Mexico_City")
        st.selectbox(
            label="Activo",
            options=machines,
            key=f"{self.__class__.__name__}_machine_selectbox",
            on_change=self._update_asset,
            index=0,
        )
        if self.edge is not None:
            st.selectbox(
                label="Dispositivo",
                options=self.current_readable_names_map.values(),
                key=f"{self.__class__.__name__}_edge_selectbox",
                on_change=self._update_edge,
                index=0,
            )
            if self.edge_kind in self._temperature_kind:
                st.radio(
                    label="\n",
                    options=(
                        self._per_last_signals,
                        self._per_time_sheet,
                    ),
                    index=1,
                    key=f"{self.__class__.__name__}_sheet_radio",
                    on_change=self._update_sheet,
                )

            with st.form("range_form"):
                st.markdown(
                    "**<p align='center'>Configura tu visualización de data</p>**",
                    unsafe_allow_html=True,
                )
                st.date_input(
                    "Introduce una fecha",
                    datetime.datetime.today()
                    if "date" not in self.data_request.range_time
                    else self.data_request.range_time["date"],
                    key=f"{self.__class__.__name__}_date_input",
                )
                st.time_input(
                    "Desde:",
                    timezone.localize(datetime.datetime.now())
                    if "start" not in self.data_request.range_time
                    else self.data_request.range_time["start"],
                    key=f"{self.__class__.__name__}_start_time_input",
                )
                st.time_input(
                    "Hasta:",
                    timezone.localize(datetime.datetime.now())
                    if "end" not in self.data_request.range_time
                    else self.data_request.range_time["end"],
                    key=f"{self.__class__.__name__}_end_time_input",
                )

                if self.edge_kind in self._images_kind:
                    st.selectbox(
                        label="Tipo de formato",
                        options=(self._animation_data_format, ),
                        index=0,
                        key=f"{self.__class__.__name__}_format_selectbox",
                    )

                elif self.edge_kind in self._vibrational_kind:
                    self.vibrational_perceptions = (
                        vibration_utils.vibrational_perceptions(
                            edges_data=self.edges_data,
                            current_machine=self.current_machine,
                            current_edge=self.edge,
                            project_data=self.app_state.project_model,
                        )
                    )
                    st.selectbox(
                        label="Dimensión de percepción",
                        options=self.vibrational_perceptions,
                        index=0,
                        key=f"{self.__class__.__name__}_perception_selectbox",
                    )
                else:
                    pass

                st.form_submit_button("Aceptar", on_click=self._update_range_time)
        else:
            pass

    def _update_edge(self):
        self.edge = st.session_state[f"{self.__class__.__name__}_edge_selectbox"]
        print(f"Current edge set to {self.edge}")
        for non_readable, readable in self.current_readable_names_map.items():
            if self.edge == readable:
                self.edge = non_readable
        self.edge_kind = self.get_kind_from_edge()

    def _update_asset(self):
        machine = st.session_state[f"{self.__class__.__name__}_machine_selectbox"]
        self.current_machine = machine
        self.edges = self.get_edges_from_asset(self.data["machines"][machine])
        self.current_readable_names_map = self.names_human_readable()
        if not self.edges:
            self.edge = None
        else:
            self.edge = self.edges[0]
            self.edge_kind = self.get_kind_from_edge()

        self.widget = None
        self._images_charts = ImagesDataCharts()
        self._vibration_charts = VibrationDataCharts()
        self._temperature_charts = TemperatureDataCharts()
        self.data_request = DataRequest()

    def _update_sheet(self):
        sheet = st.session_state[f"{self.__class__.__name__}_sheet_radio"]
        self.sheet = sheet

    def _update_perceptions(self):
        perception = st.session_state[f"{self.__class__.__name__}_perception_selectbox"]
        self.current_perception = perception

    def _update_range_time(self):
        self._update_edge()
        self._redraw_required = True
        if self.edge_kind not in self._vibrational_kind:
            data_to_render = {}
            perceptions = self.get_perceptions_for_robot_devices()
            for perception in perceptions:
                self.data_request = DataRequest(
                    range_time={
                        "date": st.session_state[
                            f"{self.__class__.__name__}_date_input"
                        ],
                        "start": st.session_state[
                            f"{self.__class__.__name__}_start_time_input"
                        ],
                        "end": st.session_state[
                            f"{self.__class__.__name__}_end_time_input"
                        ],
                    },
                    edge_kind=self.edge_kind,
                    bigquery_dataset=self._bigquery_dataset,
                    tenant_id=self._tenant_id,
                    edge=self.edge,
                    perception=perception,
                    bqclient=self.bqclient,
                )
                time_data = self.data_request.get_data_from_table_in_dataset("time", self.app_state.authentication_info.user_role)
                data_to_render.update({perception: time_data})
            self.formatted_asset_name = self.formatting_readable_strings(
                self.current_machine
            )
            if self.edge_kind in self._images_kind:
                self.data_format = st.session_state[
                    f"{self.__class__.__name__}_format_selectbox"
                ]
                self.normal_data = data_to_render[self._image_perception]
                self.thermographic_data = data_to_render[self._thermographic_perception]
                data_to_annotate = self.normal_data
            elif self.edge_kind in self._complex_kind:
                self.normal_data = data_to_render[self._image_perception]
                self.thermographic_data = data_to_render[self._thermographic_perception]
                self.normal_data = data_to_render[self._image_perception]

                self.temperature_data = complex_utils._get_temperatures(
                    data_to_render[self._temperature_perception]
                )
                data_to_annotate = self.normal_data
            elif self.edge_kind in self._temperature_kind:
                if self.sheet == self._per_last_signals:
                    data_last_temperature_values = (
                        temperature_utils.get_last_temperature_values(
                            self._tenant_id,
                            f"{self._bigquery_dataset}.time",
                            self.edge,
                            self.current_perception,
                            self.data_request.bqclient
                        )
                    )
                    self.last_temperatures_values = (
                        data_last_temperature_values.values()
                    )
                    self.last_temperatures_dates = data_last_temperature_values.keys()
                    data_to_annotate = None
                else:
                    self.temperature_data = complex_utils._get_temperatures(
                        data_to_render[self._temperature_perception]
                    )
                    data_to_annotate = data_to_render[self._temperature_perception]
        else:
            perception = st.session_state[
                f"{self.__class__.__name__}_perception_selectbox"
            ]
            self.current_perception = perception
            self.data_request = DataRequest(
                range_time={
                    "date": st.session_state[f"{self.__class__.__name__}_date_input"],
                    "start": st.session_state[
                        f"{self.__class__.__name__}_start_time_input"
                    ],
                    "end": st.session_state[
                        f"{self.__class__.__name__}_end_time_input"
                    ],
                },
                edge_kind=self.edge_kind,
                bigquery_dataset=self._bigquery_dataset,
                tenant_id=self._tenant_id,
                edge=self.edge,
                perception=perception,
                bqclient=self.bqclient,
            )
            if self.current_perception == self._temperature_perception:
                temperature_signals = self.data_request.get_data_from_table_in_dataset(
                    "time",  self.app_state.authentication_info.user_role
                )
                self.temperature_data = complex_utils._get_temperatures(
                    temperature_signals
                )
                data_to_annotate = temperature_signals
            else:
                self.vibration_data = self.data_request.get_data_from_table_in_dataset(
                    "time",  self.app_state.authentication_info.user_role
                )
                self.fft_metrics_df = (
                    self.data_request._get_data_from_table_in_dataset_ordered_by_value(
                        "fft_metrics"
                    )
                )
                self.psd_metrics_df = (
                    self.data_request._get_data_from_table_in_dataset_ordered_by_value(
                        "psd_metrics"
                    )
                )
                data_to_annotate = self.vibration_data

        if (
            data_to_annotate
            and self.app_state.authentication_info.user_role == AuthenticationInfo.UserRole.Admin
            and len(self.data['machines'][self.current_machine]['tasks'].keys())
        ):
            asset_tasks = self.data['machines'][self.current_machine]['tasks']
            labels = []
            for task in asset_tasks.keys():
                if self.edge in self.data['machines'][self.current_machine]['tasks'][task]['edges']:
                    labels = self.data['machines'][self.current_machine]['tasks'][task]['label_def']['labels']
            if labels:
                self.widget = AnnotationWidget(
                    labels,
                    self.edges_data[self.current_machine],
                    data_to_annotate,
                    self._bigquery_dataset,
                    self.data_request.bqclient,
                )
        else:
            self.widget = None

    def _render_content(self):
        log.info("Starting rendering of Analytics...")
        self.get_data()
        annotation_row = None
        self.formatted_asset_name = self.formatting_readable_strings(
            self.current_machine
        )
        padding = 0
        st.markdown(
            """
        <style>
        .big-font {
            font-size:8px !important;
        }
        </style>
        """,
            unsafe_allow_html=True,
        )
        st.markdown(
            f""" <style>
                                                    .reportview-container .main .block-container{{
                                                        padding-bottom: {padding}rem;
                                                    }} </style> """,
            unsafe_allow_html=True,
        )
        if self.edge is None:
            st.header(f"{self.formatted_asset_name}")
            st.info(
                f"No hay dispositivos monitoreando en {self.formatted_asset_name}"
            )
        else:
            if self.edge_kind in self._images_kind:
                c = st.empty()
                with c.container():
                    st.header(f"Análisis de imagenes")
                    st.subheader(f"{self.formatted_asset_name}")
                    c = st.empty()
                    if "end" not in self.data_request.range_time:
                        with c.container():
                            st.success(
                                "Introduce un intervalo de tiempo para visualizar los datos."
                            )
                    else:
                        annotation_row = st.empty()
                        if self._redraw_required:
                            computer_vision_utils.get_images_analysis(
                                user=self.app_state.authentication_info.user_role,
                                normal_data=self.normal_data,
                                thermographic_data=self.thermographic_data,
                                range_time=self.data_request.range_time,
                                data_format=self.data_format,
                                images_data_format=self._images_data_format,
                                timages_data_format=self._timages_data_format,
                                animation_data_format=self._animation_data_format,
                                images_charts=self._images_charts
                            )
                            self._redraw_required = False

                        self._draw_images_data()

            elif self.edge_kind in self._vibrational_kind:
                if self.current_perception != self._temperature_perception:
                    st.header(f"Análisis de vibraciones")
                    st.subheader(f"{self.sheet}: {self.formatted_asset_name}")
                    c = st.empty()
                    if "end" not in self.data_request.range_time:
                        with c.container():
                            st.success(
                                "Introduce un intervalo de tiempo para visualizar los datos."
                            )
                    else:
                        with c.container():
                            st.markdown(f"##### {self.current_perception}")
                            annotation_row = st.empty()
                            if self._redraw_required:
                                vibration_utils.plotting_time_data(
                                    self.app_state.project_model,
                                    self.current_perception,
                                    self.vibration_data,
                                    self.fft_metrics_df,
                                    self.psd_metrics_df,
                                    self.app_state.authentication_info.user_role,
                                    self._vibration_charts
                                )
                                self._redraw_required = False
                            self._draw_vibrational_data()

                else:
                    st.header(f"Análisis de temperatura")
                    st.subheader(f"{self.sheet}: {self.formatted_asset_name}")
                    if self.sheet == self._per_last_signals:
                        if self.last_temperatures_values is None:
                            st.info(f"Sin registro en la última hora.")
                        else:
                            st.markdown(
                                f"#### Temperatura actual de {self.formatted_asset_name}"
                            )
                            temperature_utils.actual_machine_temperature(
                                self.last_temperatures_values
                            )
                            st.markdown("\n")
                            st.markdown("\n")
                            if self._redraw_required:
                                self._temperature_charts.temp_chart = temperature_utils.plot_asset_temperatures(
                                    temperature_values=self.last_temperatures_values,
                                    temperature_dates=self.last_temperatures_dates,
                                    name="Temperatura",
                                )
                                self._redraw_required = False
                            self._draw_temperature_data()
                    else:
                        c = st.empty()
                        if "end" not in self.data_request.range_time:
                            with c.container():
                                st.success(
                                    "Introduzca un intervalo de tiempo para visualizar los datos."
                                )
                        else:
                            with c.container():
                                if not self.temperature_data[0] is None:
                                    annotation_row = st.empty()
                                    if self._redraw_required:
                                        self._temperature_charts.temp_chart = temperature_utils.plot_asset_temperatures(
                                            temperature_values=self.temperature_data[0],
                                            temperature_dates=self.temperature_data[1],
                                            name=f"Temperatura de {self.current_machine}",
                                        )
                                        self._redraw_required = False

                                    self._draw_temperature_data()

                                else:
                                    st.text("No hay data en ese intervalo")
            elif self.edge_kind in self._complex_kind:
                st.header("Análisis compuesto")
                st.subheader(f"{self.formatted_asset_name}")
                c = st.empty()
                if "end" not in self.data_request.range_time:
                    with c.container():
                        st.success(
                            "Introduce un intervalo de tiempo para visualizar los datos."
                        )
                else:
                    annotation_row = st.empty()
                    if self._redraw_required:
                        self._redraw_required = False
                        complex_utils.get_images_and_temperature_analysis(
                            user=self.app_state.authentication_info.user_role,
                            normal_data=self.normal_data,
                            thermographic_data=self.thermographic_data,
                            range_time=self.data_request.range_time,
                            data_format=self.data_format,
                            images_data_format=self._images_data_format,
                            timages_data_format=self._timages_data_format,
                            animation_data_format=self._animation_data_format,
                            temperatures_values=self.temperature_data[0],
                            temperatures_dates=self.temperature_data[1],
                            current_machine=self.current_machine,
                            images_charts=self._images_charts,
                            temperature_charts=self._temperature_charts
                        )
                    self._draw_temperature_data()
                    self._draw_images_data()

            else:
                st.error("Escoge otro asset para visualizar sus datos")
            if (
                annotation_row
                and self.app_state.authentication_info.user_role
                == AuthenticationInfo.UserRole.Admin
            ):
                with annotation_row:
                    if self.widget:
                        self.widget._render()

    def _draw_vibrational_data(self):
        if not self._vibration_charts.fft_metrics_df is None:
            st.markdown("#### FFT")
            st.plotly_chart(self._vibration_charts.fft_chart, use_container_width=True)
            st.markdown("#### Métricas de la FFT")
            vibration_utils.metrics_visualization(df=self._vibration_charts.fft_metrics_df)
            st.markdown("***")

        if not self._vibration_charts.psd_metrics_df is None:
            st.markdown("#### PSD")
            st.plotly_chart(self._vibration_charts.psd_chart, use_container_width=True)
            st.markdown("#### Métricas de la PSD")
            vibration_utils.metrics_visualization(df=self._vibration_charts.psd_metrics_df)

    def _draw_images_data(self):
        # draw normal images
        log.info("Start to draw images on browser...")
        st.markdown(f"#### Imágenes fotográficas")
        global _IMAGES_RANGE_TIME_EXCEEDED

        if _IMAGES_RANGE_TIME_EXCEEDED:
            st.warning("La cantidad de imágenes encontradas es muy grande. Ingrese un intervalo menor "
                     "para visualizar las imágenes.")
            return

        if self._images_charts.normal_imgs_chat:
            st.plotly_chart(self._images_charts.normal_imgs_chat,
                            use_container_width=True)
        else:
            st.info(
                f"No hay data para mostrar en {self.data_request.range_time['date'].strftime('%Y/%m/%d')} "
                f"de las {self.data_request.range_time['start'].strftime('%H:%M')} a "
                f"las {self.data_request.range_time['end'].strftime('%H:%M')}"
            )

        st.markdown("#### Imágenes termográficas")
        if self._images_charts.thermographic_imgs_datacharts:
            st.plotly_chart(self._images_charts.thermographic_imgs_datacharts,
                            use_container_width=True)
        else:
            st.info(
                f"No hay data para mostrar en {self.data_request.range_time['date'].strftime('%Y/%m/%d')} "
                f"de las {self.data_request.range_time['start'].strftime('%H:%M')} a "
                f"las {self.data_request.range_time['end'].strftime('%H:%M')}"
            )

    def _draw_temperature_data(self):
        log.info("Start to draw images on browser...")

        st.subheader("Temperaturas")
        if self._temperature_charts.temp_chart:
            st.plotly_chart(self._temperature_charts.temp_chart,
                            use_container_width=True)
        else:
            st.text("No se consiguió datos para graficar temperatura")



import time
import logging
import threading

class NullHandler(logging.Handler):
    def emit(self, record):
        pass
log = logging.getLogger('RateCalculator')
log.setLevel(logging.INFO)
log.addHandler(NullHandler())

DEFAULT_TSBUF_SIZE = 10

class RateCalculatorError(Exception):
    '''
    \brief Exception class associated with the rate calculator.
    '''
    
    NOT_ENOUGH_DATA          = 1
    
    descriptions = { 
        NOT_ENOUGH_DATA:     'not enough data to calculate rate', 
    }
    
    def __init__(self,errorCode,details=None):
        self.errorCode  = errorCode
        self.details    = details
    
    def __str__(self):
        try:
            output = self.descriptions[self.errorCode]
            if self.details:
                output += ': ' + str(self.details)
            return output
        except KeyError:
            return "Unknown error: #" + str(self.errorCode)

class RateCalculator(object):
    '''
    \brief Rate calculator class.
    '''
    
    def __init__(self, tsBufSize=DEFAULT_TSBUF_SIZE):
        
        # log
        log.info("creating instance")
        
        # store params
        self.tsBufSize = tsBufSize
        
        # local variables
        self.tsBufLock = threading.Lock()
        self.tsBuf     = []
    
    def signalEvent(self,ts=None):
        
        if not ts:
            ts = time.time()
        
        # log
        log.debug("adding event at {0}".format(ts))
        
        with self.tsBufLock:
            self.tsBuf.append(ts)
            while len(self.tsBuf)>self.tsBufSize:
                self.tsBuf.pop(0)
    
    def clearBuf(self) :
        with self.tsBufLock:
            self.tsBuf = []
        
    def getRate(self):
        
        returnVal = None
        
        with self.tsBufLock:
            if  len(self.tsBuf)>self.tsBufSize:
                raise SystemError("rate buffer has {0} elements, expected at most {1}".format(len(self.tsBuf),self.tsBufSize))
            if len(self.tsBuf)==self.tsBufSize:
                returnVal = float(self.tsBufSize-1)/(self.tsBuf[-1]-self.tsBuf[0])
        
        if not returnVal:
            raise RateCalculatorError(RateCalculatorError.NOT_ENOUGH_DATA)
        else:
            return returnVal


import threading
from txp.common.configuration.job import JobTask
from txp.common.edge import EdgeDescriptor
from txp.devices.devices_manager import DevicesManager
from datetime import datetime
from txp.common.configuration import datetime_utils
import time
from typing import List, Dict, Tuple, Set
from math import gcd
from functools import reduce
from txp.common.config import settings
import copy
import enum
import logging

from txp.devices.sampling_task import SamplingTask

log = logging.getLogger(__name__)
log.setLevel(settings.txp.general_log_level)


class _WindowPolicyEnum(enum.Enum):
    """This policy is used by the WindowPolicyManager to decide what actions
    to take when an edge doesn't reports the expected accomplished clock pulses.

    Below there's an explanation of the policies:
        - no_wait: The Asset observation clock will not do anything about a delay of
            a driver. It will only log the WARNING messages.
        - wait_x_time: The observation clock will wait for the driver to catch up
            given a configuration timeout.
    """

    NO_WAIT = "no_wait"
    WAIT_X_TIME = "wait_x_time"


class WindowPolicyManager(threading.Thread):
    """This WindowPolicyManager conceptualizes the unit of control to check the drivers
    sampling cadence within the sampling windows of a job task.

    The Manager knows about:
        - The total number of sampling windows to execute, given the job task start time
            and end time.
        - The total number of clocks to execute in each of the mentioned sampling windows
            to execute.

    Based on those values, the manager will check and determine what policies to
    apply to the sampling cadence of the drivers.
    """
    def __init__(
        self,
        identifier: str,
        configuration_id: str,
        tenant_id: str,
        asset_task: JobTask,
        edges_logical_ids: List[str],
        devices_manager: DevicesManager,
        cancel_sampling: threading.Event,
    ):
        """
        Args:
            configuration_id: The configuration id received by the Gateway.
            asset_task: The JobTask to execute in this WindowPolicyManager.
            edges_logical_ids: The list of drivers logical IDs involved in the task
                to observe the asset.
            devices_manager: The DevicesManager singleton used in the system.
        """
        super(WindowPolicyManager, self).__init__(name=identifier, daemon=True)

        # Task context
        self._configuration_id = configuration_id
        self._tenant_id = tenant_id
        self._task: JobTask = asset_task
        self._logical_ids: List[str] = edges_logical_ids
        self._devices_manager_service: DevicesManager = devices_manager
        self._logical_ids: List[str] = [
            logical_id
            for logical_id in edges_logical_ids
            if not self._devices_manager_service.is_actuator_device(logical_id)
        ]
        self._clock_pulse_value: int = self._get_manager_clock_value()
        self._canceled: threading.Event = cancel_sampling
        self._gateway_task_id: int = 0

        # Control attributes
        self._current_clock: int = 0
        self._current_window: int = 0
        self._clocks_per_window: int = self._get_total_clocks_per_window()

        current_time = datetime.now().time()

        # If the task was already "started", update to the current time
        if (
            current_time.hour != self._task.parameters.start_time.hour or
            current_time.minute != self._task.parameters.start_time.minute
        ):
            self._task.parameters.start_time = current_time

        self._num_windows: int = datetime_utils.get_num_windows_in_task(
            self._task.parameters
        )

        if self._num_windows < 0:
            log.error(
                f"There was an error computing the number of "
                f"sampling windows: {self._num_windows}"
            )
            log.warning(
                f"Asset Observation Clock set to perform 0 sampling windows after error"
            )

        elif self._num_windows == 0:
            log.warning(
                f"Total number of sampling windows in task is zero(0)."
                f"This might be caused by a configuration error. Check "
                f"the driver task configuration"
            )

        self._window_start_time: datetime = None
        self._dead_time_counter: int = 0

        # Stats attributes used in unit testing
        self._total_num_pulses_executed: int = 0

        # Policies Handling initialization
        self._policies: Dict = {
            _WindowPolicyEnum.NO_WAIT: self._no_wait_policy_apply,
            _WindowPolicyEnum.WAIT_X_TIME: self._wait_x_time_policy_apply,
        }

        try:
            self._policy = _WindowPolicyEnum(
                settings.gateway.sampling.default_observation_policy
            )
            log.debug(
                f"{self.__class__.__name__} policy configured to: {self._policy.name}"
            )
        except:
            log.warning(
                f"{self.__class__.__name__} could not parse the settings policy. "
                f"Fallback policy to {_WindowPolicyEnum.NO_WAIT.name}"
            )
            self._policy = _WindowPolicyEnum.NO_WAIT

    def _get_manager_clock_value(self) -> int:
        """Returns the clock pulse value to base the synchronisation of this sampling window
        manager.

        The clock pulse value is the greatest common divisor of all the members


        The clock pulse value determines the period of time used by this SamplingWindowManager
        to execute control policies on the sampling cadence of the drivers used in the job task.
        """
        drivers_clock_pulses = [
            self._devices_manager_service.get_driver_observation_pulse(logical_id)
            for logical_id in self._logical_ids
        ]

        gcd_value = reduce(gcd, drivers_clock_pulses)

        return gcd_value

    def _get_total_clocks_per_window(self) -> int:
        """Returns the total clocks to be executed in each one of the
        sampling windows."""
        total_clock_pulses = int(
            self._task.parameters.sampling_window.sampling_time
            / self._clock_pulse_value
        )

        if not total_clock_pulses:
            log.error(
                f"{self.__class__.name} clock has 0 ticks to execute. Sampling time is less than minimal"
                " drivers observation pulse."
            )

        return total_clock_pulses

    def _apply_clock_control(self):
        """Returns the expected packages to be received in the clock pulse

        When this method gets called, each driver will have its own account
        of expected packages given by the equation:
            expected_packages = (clock value * current clock) / driver clock value.

        This equation allows the control to be relative, and it helps to deal with
        real world time which is shared by the operative system on its own.
        """
        completed_clocks_per_driver = self._get_completed_clocks_per_drivers()
        delayed_drivers: List[str] = []

        for logical_id in self._logical_ids:
            expected_packages = self._get_num_expected_packages(logical_id)

            if (
                self._devices_manager_service.is_virtual_driver(logical_id)
                and self._current_clock != self._clocks_per_window
            ):
                log.debug(
                    "Edge is virtual. Virtual edges are checked at the end of the clock."
                )
                continue  # process next edge

            if (
                expected_packages != completed_clocks_per_driver[logical_id]
            ):  # The driver is late. Adding Driver to delayed drivers.
                delayed_drivers.append(logical_id)
                log.debug(f"Driver {logical_id} is delayed. "
                          f"Expected clocks: {expected_packages} - "
                          f"Received clocks: {completed_clocks_per_driver[logical_id]}")

            else:
                log.debug(
                    f"Driver {logical_id} is synchronized with the expected pulses."
                    f"Expected pulses: {expected_packages} - "
                    f"Received pulses: {completed_clocks_per_driver[logical_id]}"
                )

        # Apply policy for the delayed drivers
        if delayed_drivers:
            log.info("Applying Windows policy to delayed drivers.")
            self._policies[self._policy](delayed_drivers)
        else:
            log.info(f"Clock pulse {self._current_clock} for window {self._current_window} "
                     f"finished without delayed drivers.")

    def _get_completed_clocks_per_drivers(self):
        return self._devices_manager_service.get_drivers_completed_clock(
            self._logical_ids
        )

    def _get_edges_to_enable_in_next_clock(self) -> List[str]:
        """Returns the list of edges logical ids to enable in the next clock pulse."""
        if self._current_clock == 1:
            edges_to_enable = self._logical_ids

        else:
            elapsed_sampling_time = self._current_clock * self._clock_pulse_value
            leftover_sampling_time = self._task.parameters.sampling_window.sampling_time - elapsed_sampling_time
            log.debug(f"Leftover time in sampling time available to enable drivers: {leftover_sampling_time}")
            edges_to_enable = [
                logical_id for logical_id in self._logical_ids if
                (
                    self._devices_manager_service.get_driver_observation_pulse(logical_id) <= leftover_sampling_time
                    and (self._clock_pulse_value * self._current_clock) %
                    self._devices_manager_service.get_driver_observation_pulse(logical_id) == 0
                    and not self._devices_manager_service.is_virtual_driver(logical_id)
                )
            ]

        log.info(f"Edges to be enabled in next clock control: {edges_to_enable}")
        return edges_to_enable

    def _get_num_expected_packages(self, edge_logical_id: str) -> int:
        """Returns the number of expected packages of the given edge
        logical id for the current clock pulse value."""
        return int(
            (self._clock_pulse_value * self._current_clock)
            / self._devices_manager_service.get_driver_observation_pulse(
                edge_logical_id
            )
        )

    def _send_drivers_tasks(self) -> None:
        """Sends the sampling tasks to the Drivers that monitors the asset"""
        tasks: List[Tuple] = []

        # Adds the observation_id to the packages sent to the edges.
        task_copy = copy.copy(self._task)
        task_copy.parameters.sampling_window.observation_timestamp = time.time_ns()
        task_copy.parameters.sampling_window.sampling_window_index = self._current_window - 1
        task_copy.parameters.sampling_window.number_of_sampling_windows = self._num_windows
        task_copy.parameters.sampling_window.gateway_task_id = self._gateway_task_id

        for logical_id in self._logical_ids:
            log.debug(
                f"Computing Sampling tasks to send to driver {logical_id} in the next "
                f"asset sampling window."
            )
            if self._devices_manager_service.is_virtual_driver(logical_id):
                log.debug(
                    f"Driver {logical_id} is virtual edge. Assuming 1 sampling task to send"
                )
                sampling_task: SamplingTask = SamplingTask(
                    self._configuration_id, task_copy.parameters,
                    self._tenant_id
                )
                tasks.append((logical_id, sampling_task))

            else:
                num_clock_pulses: int = int(
                    self._task.parameters.sampling_window.sampling_time
                    / self._devices_manager_service.get_driver_observation_pulse(
                        logical_id
                    )
                )
                driver_task_list = [
                    (
                        logical_id,
                        SamplingTask(self._configuration_id, task_copy.parameters, self._tenant_id),
                    )
                ]
                driver_tasks_list = driver_task_list * num_clock_pulses
                log.debug(
                    f"Driver {logical_id} will receive {len(driver_tasks_list)} sampling tasks"
                )
                tasks = tasks + driver_tasks_list

        self._devices_manager_service.receive_sampling_tasks_list(tasks)

    def run(self) -> None:
        # Clock control loop for the Task Sampling windows
        log.info(
            f"{self.__class__.name} clock for {self._task.machines} has {self._clocks_per_window} pulses "
            f"to execute per window"
        )
        log.info(f"Total number of sampling windows in task {self._num_windows}")

        log.info(
            f"{self.__class__.name} for {self._task.machines} has a minimal "
            f"clock pulse of {self._clock_pulse_value}"
        )

        self._gateway_task_id = time.time_ns()

        while True:
            if self._current_window == self._num_windows:
                log.info(f"{self.__class__.name} clock completed its task.")
                break

            self._current_window += 1
            self._current_clock = 0
            self._dead_time_counter = 0
            self._send_drivers_tasks()
            self._devices_manager_service.start_sampling_windows(self._logical_ids)
            self._window_start_time = datetime.now()
            log.debug(f"New clock execution started at {self._window_start_time}")
            self._devices_manager_service.clear_drivers_completed_clocks(
                self._logical_ids
            )

            while True:
                if self._current_clock == self._clocks_per_window:
                    log.info(
                        "Sampling Window execution completed successfully by observation clock"
                    )
                    break

                self._current_clock += 1
                log.info(
                    f"Window manager will execute the next clock pulse: {self._current_clock}"
                )
                self._devices_manager_service.enable_drivers_clock_pulse(
                    self._get_edges_to_enable_in_next_clock()
                )

                clock_pulse_start = datetime.now()
                while (
                    datetime.now() - clock_pulse_start
                ).total_seconds() < self._clock_pulse_value:
                    time.sleep(5)
                    if self._canceled.is_set():
                        log.debug(
                            "Work task canceled because the cancel event was triggered."
                        )
                        self._devices_manager_service.stop_sampling_windows(
                            self._logical_ids
                        )
                        return

                self._apply_clock_control()
                self._total_num_pulses_executed += 1

                self._devices_manager_service.get_drivers_completed_clock(
                    self._logical_ids
                )

            log.debug(f"Sampling Window dead time execution started")
            self._devices_manager_service.notify_sampling_window_completed(self._logical_ids)
            remaining_dead_time = (
                self._task.parameters.sampling_window.dead_time
                - self._dead_time_counter
            )
            if remaining_dead_time < 0:
                log.warning(
                    f"Dead time was exceeded by the policy by {remaining_dead_time} secs. "
                    f"The system configuration should "
                    "be reviewed"
                )
            else:
                log.debug(
                    f"Window Dead time remaining time to wait: {remaining_dead_time}"
                )
                clock_pulse_start = datetime.now()
                while (datetime.now() - clock_pulse_start).total_seconds() < abs(
                    remaining_dead_time
                ):
                    time.sleep(5)
                    if self._canceled.is_set():
                        log.debug(
                            "Work task canceled because the cancel event was triggered."
                        )
                        break

            self._devices_manager_service.stop_sampling_windows(
                self._logical_ids
            )
            log.info("Sampling Window dead time execution completed successfully")

    def _no_wait_policy_apply(self, logical_ids: List[str]) -> None:
        """Applies the no_wait policy for the reported driver"""
        for logical_id in logical_ids:
            log.warning(
                f"Driver {logical_id} was reported as delayed, but configured policy "
                f"don't do nothing."
            )

    def _wait_x_time_policy_apply(self, logical_ids: List[str]) -> None:
        """Applies the wait_x_time policy for the reported driver.

        It will block the current thread for the configured value of x.
        If the driver catch up, then it will return immediately.
        """
        x = settings.gateway.sampling.wait_x_time_seconds_interval
        start = datetime.now()
        time.sleep(x)
        completed_clocks_per_driver = (
            self._devices_manager_service.get_drivers_completed_clock(self._logical_ids)
        )
        end = datetime.now()
        elapsed = (end - start).total_seconds()
        self._dead_time_counter += elapsed

        for logical_id in logical_ids:
            expected_num_samplings = self._get_num_expected_packages(logical_id)

            if expected_num_samplings == completed_clocks_per_driver[logical_id]:
                log.info(
                    f"Driver {logical_id} did catch up after policy: "
                    f"Expected clocks: {expected_num_samplings} - "
                    f"Reported clocks: {completed_clocks_per_driver[logical_id]}"
                )

            else:
                log.info(
                    f"The driver {logical_id} did not catch up after policy. Requesting the driver"
                    f"to stop its collection."
                )
                self._devices_manager_service.disable_drivers_clock_pulse([logical_id])

        # TODO: policies dead time count could be handled elegantly with a decorator
        log.debug(
            f"Applying policy {_WindowPolicyEnum.WAIT_X_TIME.name} "
            f"took {elapsed} from windows dead time"
        )



import json
import ray
import logging
from google.oauth2 import service_account
import google.cloud.firestore as firestore
import google.cloud.bigquery as bigquery
import txp.cloud.controllers.bigquery as bq
from txp.common.ml.annotation import AnnotationLabelPayload
from typing import List, Set, Dict
from google.cloud.exceptions import NotFound
from txp.ml.training_service.dataset_committer import DatasetCommitPayload
from txp import cloud as cloud_utils
import threading
import time


@ray.remote
class AnnotationManager:
    """This class is in charge of processing anything that arrives at the
    /annotation endpoint.

    The responbilitites for this class are:
        - Processing incoming request of annotations objects to store
            in the persistence layer.
        - Asynchronously write annotation objects from the persistence layer
            into the BigQuery tables.
        - Answer if there are pending annotations to be processed for a given
            task.
    """

    def __init__(self, credentials_str, log_level=logging.INFO):
        self._annotations_collection_name: str = "tasks_annotations"
        self._tables_names: List[str] = [
            "time",
            "fft",
            "psd",
            "time_metrics",
            "fft_metrics",
            "psd_metrics",
        ]
        self._log_level = log_level
        logging.basicConfig(level=log_level)
        json_dict_service_account = json.loads(credentials_str, strict=False)
        credentials = service_account.Credentials.from_service_account_info(
            json_dict_service_account
        )
        self._credentials = credentials
        self._db_lock: threading.Lock = threading.Lock()
        self._annotations_rate: int = 50
        self._annotations_thread: threading.Thread = threading.Thread(
            daemon=True,
            name="annotations_processing_thread",
            target=self._process_background_annotations,
        )
        self._annotations_thread.start()
        self._current_pending_annotations: Set[str] = set()
        logging.info(f"{self.__class__.__name__} Created successfully.")

    def process_new_annotations(
        self, annotations: List[AnnotationLabelPayload]
    ) -> bool:
        logging.info("Processing new annotations received.")
        logging.info(f"Annotations to process {annotations.__len__()}")
        db: firestore.Client = firestore.Client(
            credentials=self._credentials, project=self._credentials.project_id
        )

        if not annotations:
            return True

        try:
            # Note: currently, we simplify code by the assumption that labelers are only
            #   labelling one edge at a time.
            self._db_lock.acquire()
            annotations_map = list(map(lambda ann: ann.dict(), annotations))

            doc_ref: firestore.DocumentReference = db.collection(
                self._annotations_collection_name
            ).document(annotations[0].tenant_id + "_" + annotations[0].edge_logical_id)

            doc: firestore.DocumentSnapshot = doc_ref.get()

            if doc.exists:
                doc.reference.update(
                    {"annotations": firestore.ArrayUnion(annotations_map)}
                )
            else:
                doc.reference.set({"annotations": annotations_map})

        except NotFound:
            logging.error(
                "Trying to add annotations to document, but the document was not found."
            )
            return False

        except Exception as e:
            logging.error(f"Unknown exception while writing firestore document: {e}")

        else:
            logging.info(
                f"Successfully updated the annotations for the document with new annotations"
            )
            return True

        finally:
            self._db_lock.release()
            db.close()

    def validate_commit(self, commit_payload: DatasetCommitPayload) -> bool:
        """Returns True if it's possible to perform a requested commit.

        The commit will be possible if all the pending annotations for the
        requested dataset commit has been written to the persistence layer.

        Returns:
            - False: the task specified in the DatasetCommitPayload was not found
                for the asset in firestore.

            - False: it's not possible to trigger the dataset commit generation
                because there are pending annotations to process.

            - True: The commit operation is valid an can be triggered
        """
        logging.info(f"Validating requested commit.")

        try:
            db: firestore.Client = firestore.Client(
                credentials=self._credentials, project=self._credentials.project_id
            )

            task = cloud_utils.get_task_from_firestore(
                db,
                commit_payload.tenant_id,
                commit_payload.machine_id,
                commit_payload.task_id,
            )
            if task is None:
                logging.error(
                    f"No task {commit_payload.task_id} was found for asset {commit_payload.machine_id} for"
                    f" tenant {commit_payload.tenant_id}"
                )
                return False
        except Exception as e:
            logging.error(f"Unexpected error while validating commit: {e}")
            return False

        # At this point, we need to validate that all the edges of the task don't have
        #   any pending annotation in the cache to be written in the persistence layer.
        for edge_id in task.edges:
            doc_ref: firestore.DocumentReference = db.collection(
                self._annotations_collection_name
            ).document(commit_payload.tenant_id + "_" + edge_id)
            annotations = doc_ref.get().get("annotations")
            if annotations:
                logging.warning(
                    "The commit isn't valid because there are pending "
                    f"annotations to process for edge {edge_id}"
                )
                return False

            if (
                commit_payload.tenant_id + "_" + edge_id
                in self._current_pending_annotations
            ):
                logging.warning(
                    "The commit isn't valid because there are annotations being"
                    f" processed for edge {edge_id}"
                )
                return False

        logging.info(f"The dataset commit is valid. There are not pending annotations.")

        return True

    def _process_background_annotations(self):
        """This infinite loop method code will pull annotations from the database
        and will update the BigQuery tables on a separated thread.

        TODO: we should handle errors in the interaction with the GCP services.
        """
        while True:
            time.sleep(60)

            logging.info(f"Pulling annotations from DB. Acquiring secure context...")
            self._db_lock.acquire()

            db: firestore.Client = firestore.Client(
                credentials=self._credentials, project=self._credentials.project_id
            )
            try:
                docs = db.collection(self._annotations_collection_name).stream()
            except Exception as e:
                logging.error(
                    f"Unexpected exception while pulling annotations from"
                    f" firestore: {e}"
                )
                continue

            finally:
                db.close()
                self._db_lock.release()

            bq_client: bigquery.Client = bigquery.Client(credentials=self._credentials)
            new_annotations: Dict[str, List[AnnotationLabelPayload]] = {}
            try:
                for doc in docs:
                    annotations = doc.get("annotations")
                    if not annotations:
                        logging.info(
                            f"Document {doc.id} doesn't have annotations. Removing"
                        )
                        doc.reference.delete()
                    else:
                        self._current_pending_annotations.add(doc.id)
                        temp_annotations = annotations[0 : self._annotations_rate]
                        logging.info(f"Taking {temp_annotations.__len__()} to process")
                        # remove taken annotations
                        doc.reference.update(
                            {"annotations": firestore.ArrayRemove(temp_annotations)}
                        )

                        new_annotations[doc.id] = list(
                            map(
                                lambda ann: AnnotationLabelPayload(**ann),
                                temp_annotations,
                            )
                        )

                if not new_annotations:
                    logging.info(f"Not annotations found to write.")
                    continue

                logging.info(f"Writing annotations into BigQuery datasets")
                for doc_id, annotations_list in new_annotations.items():
                    logging.info(f"Processing a new annotation for collection {doc_id}")
                    for annotation in annotations_list:
                        for table in self._tables_names:
                            bq.annotate_signal(
                                annotation.tenant_id,
                                f"{annotation.dataset_name}.{table}",
                                annotation.edge_logical_id,
                                annotation.observation_timestamp,
                                annotation.label.json(),
                                annotation.version,
                                bq_client,
                            )

                    self._current_pending_annotations.remove(doc_id)

            except Exception as ex:
                logging.error(
                    f"Unexpected exception while pulling annotations from"
                    f" firestore: {ex}"
                )
                continue

            finally:
                bq_client.close()



"""
This module defines the helper function to draw the package collection
panel in the browser.
"""
# =====================imports=============================================
import streamlit as st
import gateway_mocking as mocking
from txp.devices.gateway_states_enums import *
from txp.common.config import settings
import txp.devices.ux.ux_states_protocol as ux_states
from txp.devices.ux.configuration_panel import render_configuration_panel
import pandas as pd


# ==================== Functions ==========================================
def _print_details(details: ux_states.UxDetails):
    if details.kind == ux_states.UxDetails.Kind.Warning:
        st.warning(details.message)

    elif details.kind == ux_states.UxDetails.Kind.Info:
        st.info(details.message)
    else:
        st.error(details.message)


def _print_table_of_devices_states():
    st.markdown("**Tabla de estado de dispositivos**")
    table = pd.DataFrame(columns=["Serial", "Estado"])
    table.set_index("Serial", inplace=True)
    for serial, state in st.session_state.gateway_state.devices_connection_table.items():
        table.loc[serial] = str(state)
    st.table(table)


def render_package_collection():
    """The Package Collection panel is used to show stats about
    package collection stats.
    """
    st.markdown("### Recolección de Paquetes")
    gateway_state = st.session_state.gateway_state

    if isinstance(gateway_state, (ux_states.UxReadyState, ux_states.UxSamplingState)):
        st.markdown(
            f"**Total de Paquetes recolectados**: {gateway_state.collected_packages}"
        )
        st.markdown(f"**Total de Paquetes enviados:** {gateway_state.sent_packages}")

        _print_table_of_devices_states()

        st.markdown("**Paquetes recolectados por dispositivos**")
        table = pd.DataFrame(columns=["Serial", "Paquetes"])
        table.set_index("Serial", inplace=True)
        for edge in gateway_state.edges:
            table.loc[edge.logical_id] = gateway_state.packages_received_by_edge[
                edge.logical_id
            ]

        st.write(table)

        _print_details(gateway_state.details)

    if isinstance(gateway_state, ux_states.UxPreparingDevicesState):
        _print_details(gateway_state.details)
        _print_table_of_devices_states()




"""
    This module provides functions and classes that helps with programmatic interaction with
    Firestore service.
"""
from typing import List, Dict
import google.cloud.firestore as firestore
from txp.common.configuration import JobConfig
import datetime
import logging

from txp.common.ml.tasks import AssetTask
from txp.cloud.common.config import settings as cloud_settings
from txp.common.config import settings
from txp.cloud.common.utils import pull_current_configuration_from_firestore
from txp.common.edge import EdgeDescriptor, MachineMetadata
from txp.common.configuration import GatewayConfig
import txp.common.configuration.config_version as txp_config_version
from txp.common.json_complex_encoder import ComplexEncoder
import pytz
import json
import enum
import traceback

log = logging.getLogger(__name__)
log.setLevel(settings.txp.general_log_level)

# =========================== Constant Data ===============================
manual_configuration_collection = (
    cloud_settings.firestore.manual_configuration_edit_collection
)
config_form_document = cloud_settings.firestore.config_form_document
machines_group_collection_name = cloud_settings.firestore.machines_group_collection
machines_collection_name = cloud_settings.firestore.machines_collection
gateways_collection_name = cloud_settings.firestore.gateways_collection
devices_collection_name = cloud_settings.firestore.devices_collection
edges_collection_name = cloud_settings.firestore.edges_collection
jobs_collections_name = cloud_settings.firestore.jobs_collection
configurations_collection_name = cloud_settings.firestore.configurations_collection
zoned_datetime_str_format = cloud_settings.firestore.datetime_zoned_format
tenants_collection = cloud_settings.firestore.tenants_collection


configuration_reference_field_path = "configuration_ref"


# =========================== Definitions ===============================
class ConfigurationSnapshotType(enum.Enum):
    PRELIMINARY = 0
    NORMAL = 1


class ProjectModel:
    """This class represents the project model downloaded from the database.

    References:
        - https://tranxpert.atlassian.net/wiki/spaces/TD/pages/345538579/Project+structure
        - https://tranxpert.atlassian.net/wiki/spaces/TD/pages/345571390/Job+Structure
    """

    def __init__(
        self,
        current_configuration: firestore.DocumentSnapshot,
        machines_groups_docs: List[firestore.DocumentSnapshot],
        machines_docs: List[firestore.DocumentSnapshot],
        gateways_docs: List[firestore.DocumentSnapshot],
        edges_docs: List[firestore.DocumentSnapshot],
        devices_docs: List[firestore.DocumentSnapshot],
        jobs_docs: List[firestore.DocumentSnapshot],
    ):
        # Attributes to hold the dictionaries or data classes from the pulled documents
        self._config_version = current_configuration.get("configuration_id")
        self._config_since = current_configuration.get("since")

        # TODO: Machines Groups requires a typed data class
        self.machines_groups_table: Dict[str, Dict] = {}
        self.machines_table: Dict[str, MachineMetadata] = {}
        self.tasks_by_asset: Dict[str, Dict[str, AssetTask]] = {}
        self.edges_table: Dict[str, EdgeDescriptor] = {}
        self.device_kinds_table: Dict[str, Dict] = {}
        self.gateways_table: Dict[str, GatewayConfig] = {}
        self.jobs_table_by_gateway: Dict[str, JobConfig] = {}
        self.edges_states_table: Dict[str, Dict] = {}  # TODO: States should be typed
        self._build_machines_groups_table(machines_groups_docs)
        self._build_machines_table(machines_docs)
        self._build_edges(edges_docs, devices_docs)
        self._build_gateways_and_jobs(gateways_docs)

    @property
    def configuration_version(self) -> str:
        return self._config_version

    def _build_machines_groups_table(self, documents: List[firestore.DocumentSnapshot]):
        self.machines_groups_table = {}
        for machine_group_doc in documents:
            machine_group = machine_group_doc.to_dict()
            machines_ids = list(
                map(
                    lambda machine_doc_ref: machine_doc_ref.get().get("machine_id"),
                    machine_group["machines"],
                )
            )
            if not machines_ids:
                log.warning(
                    f"Machines Group {machine_group['name']} without machines "
                    f"in pulled configuration"
                )
            machine_group.update({"machines": machines_ids})
            machine_group.pop(configuration_reference_field_path)
            self.machines_groups_table[machine_group["name"]] = machine_group

        log.info(
            f"Pulled {len(self.machines_groups_table)} machines groups in configuration"
        )

    def _build_machines_table(self, documents: List[firestore.DocumentSnapshot]):
        for machine_doc in documents:
            machine = machine_doc.to_dict()
            edges_ids = list(
                map(
                    lambda edge_ref: edge_ref.get().get("logical_id"),
                    machine["associated_with_edges"],
                )
            )
            if not edges_ids:
                log.warning(
                    f"Machine {machine['machine_id']} without edges "
                    f"in pulled configuration"
                )
            machine.update({"edges_ids": edges_ids})
            self.machines_table[
                machine["machine_id"]
            ] = MachineMetadata.build_from_dict(machine)
            self.tasks_by_asset[machine['machine_id']] = {}

            for task_id, body in machine['tasks'].items():
                body['machine_id'] = machine['machine_id']
                asset = AssetTask(**body)
                self.tasks_by_asset[machine['machine_id']][body['task_id']] = asset

            if not machine["machine_id"]:
                log.error(
                    "Could not parse received MachineMetadata for"
                    f"machine {machine['machine_id']}"
                )

        log.info(f"Pulled {len(self.machines_table)} machines in configuration")

    def _build_edges(
        self,
        edges_docs: List[firestore.DocumentSnapshot],
        devices_docs: List[firestore.DocumentSnapshot],
    ):

        for device in devices_docs:
            self.device_kinds_table[device.to_dict()["kind"]] = device.to_dict()

        for edge in edges_docs:
            device_kind = edge.to_dict()["device_kind"]
            device = self.device_kinds_table[device_kind]

            if "type" in edge.to_dict():
                device_type = edge.get("type")
            elif "device_type" in edge.to_dict():
                device_type = edge.get("device_type")
            else:
                device_type = device["type"]

            edge_parameters: Dict = {
                **device["parameters"],
                **edge.to_dict()["parameters"],
            }
            edge_perceptions: Dict = device["perceptions"]
            init_dict = {
                "logical_id": edge.to_dict()["logical_id"],
                "device_kind": device_kind,
                "device_type": device_type,
                "edge_parameters": edge_parameters,
                "perceptions": edge_perceptions,
            }
            edge_descriptor = EdgeDescriptor.build_from_dict(init_dict)
            self.edges_table[edge_descriptor.logical_id] = edge_descriptor

            edge_state = edge.to_dict().get('state', None)
            self.edges_states_table[edge_descriptor.logical_id] = edge_state

        log.info(f"Pulled {len(self.edges_table)} edges from project model")

    def _build_gateways_and_jobs(self, gateways_docs: List[firestore.DocumentSnapshot]):
        for gateway_doc in gateways_docs:
            gateway_dict = gateway_doc.to_dict()
            machines_ids: List[str] = list(
                map(
                    lambda machine_doc_ref: machine_doc_ref.get().get("machine_id"),
                    gateway_dict["machines"],
                )
            )
            gateway_dict.update({"machines_ids": machines_ids})
            gateway_config: GatewayConfig = GatewayConfig.build_from_dict(gateway_dict)
            if not gateway_config:
                log.error(
                    f"Error parsing GatewayConfig for gateway {gateway_dict['gateway_id']}"
                )
                continue
            job: firestore.DocumentSnapshot = gateway_doc.get("has_job").get()
            job_config: JobConfig = JobConfig.build_from_dict(job.to_dict())
            if not job_config:
                log.error(
                    f"Error parsing JobConfig for gateway {gateway_dict['gateway_id']}"
                )

            self.jobs_table_by_gateway[gateway_config.gateway_id] = job_config
            self.gateways_table[gateway_config.gateway_id] = gateway_config

        log.info(f"Pulled {len(self.gateways_table)} gateways from configuration")


# =========================== Useful Functions ===============================
def _pull_docs_from_collection_associated_with_configuration(
    db: firestore.Client,
    collection_name: str,
    configuration_ref: firestore.DocumentReference,
    as_dict: bool = True,
) -> List[firestore.DocumentSnapshot]:
    """Returns the documents of a collection that has the configuration value equal to
    the configuration reference received.

    Returns:
        The list of documents found for the query.
    """
    documents = (
        db.collection(collection_name)
        .where(configuration_reference_field_path, "==", configuration_ref)
        .get()
    )

    if not documents:
        log.warning(
            f"No documents found in collection {collection_name} for configuration {configuration_ref.id}"
        )

    if as_dict:
        documents = list(map(lambda doc: doc.to_dict(), documents))

    return documents


def pull_project_data_by_date(
    db: firestore.Client, selected_date: datetime.date
) -> "ProjectData":
    """Pulls and returns a project configuration given a date.

    TODO: A Dataclass for the project data should be added in TXP Common
    TODO: If multiple configuration were written on the same day, the return type should
        be a List of that dataclass.
    TODO: Multigateway configuration will pull multiple jobs

    Args:
        db: A configured firestore client instance.
        selected_date: The selected date to perform the query.
    """
    configuration = (
        db.collection(configurations_collection_name)
        .where("since", "<", selected_date)
        .order_by("since", direction=firestore.Query.DESCENDING)
        .limit(1)
        .get()
    )

    if not configuration:
        log.warning(f"No Configuration document was found for date: {selected_date}.")
        return None

    configuration: firestore.DocumentSnapshot = configuration[0]

    machines_groups_documents = (
        _pull_docs_from_collection_associated_with_configuration(
            db, machines_group_collection_name, configuration.reference
        )
    )

    if not machines_groups_documents:
        log.warning(
            "No Machines Groups documents were found associated "
            f"with configuration {configuration.to_dict()['configuration_id']}"
        )
        return None

    machines_documents = _pull_docs_from_collection_associated_with_configuration(
        db, machines_collection_name, configuration.reference
    )

    if not machines_documents:
        log.warning(
            f"No Machines documents were found associated "
            f"with configuration {configuration.to_dict()['configuration_id']}"
        )
        return None

    gateways_documents = _pull_docs_from_collection_associated_with_configuration(
        db, gateways_collection_name, configuration.reference
    )

    if not gateways_documents:
        log.error(
            "No Gateways documents were found associated "
            f"with configuration {configuration.to_dict()['configuration_id']}"
        )
        return None

    edges_documents = _pull_docs_from_collection_associated_with_configuration(
        db, edges_collection_name, configuration.reference
    )

    if not edges_documents:
        log.error(
            "No Edges documents were found associated "
            f"with configuration {configuration.to_dict()['configuration_id']}"
        )
        return None

    devices_documents = _pull_docs_from_collection_associated_with_configuration(
        db, devices_collection_name, configuration.reference
    )

    if not devices_documents:
        log.error(
            "No Devices documents were found associated "
            f"with configuration {configuration.to_dict()['configuration_id']}"
        )
        return None

    job_documents = _pull_docs_from_collection_associated_with_configuration(
        db, jobs_collections_name, configuration.reference
    )

    if not job_documents:
        log.error(
            "No Job documents were found associated with "
            f"with configuration {configuration.to_dict()['configuration_id']}"
        )
        return None

    job_configs: List[JobConfig] = list(
        map(lambda job_dict: JobConfig.build_from_dict(job_dict), job_documents)
    )

    if not any(job_configs):
        log.error(
            "Some/All Job Configurations obtained from Firestore could not be parsed"
        )
        return None

    ret = {
        "last_configuration": configuration,
        "machines_groups": machines_groups_documents,
        "machines": machines_documents,
        "gateways": gateways_documents,
        "edges": edges_documents,
        "devices": devices_documents,
        "jobs": job_documents,
        "query_data": {},
    }

    return ret


def get_authenticated_client(service_account_credentials) -> firestore.Client:
    """Returns a firestore.Client instance, authenticated with the provided
    credentials.

    Args:
        service_account_credentials: A service_account.Credentials object with the
            appropriate permissions to use Firestore.
    """
    return firestore.Client(
        credentials=service_account_credentials,
        project=service_account_credentials.project_id,
    )


def pull_recent_project_data(db: firestore.Client, tenant_id: str) -> "ProjectData":
    """Pulls and returns the project data from the current configuration.

    TODO: A Dataclass for the project data should be added in TXP Common
    TODO: If multiple configuration were written on the same day, the return type should
        be a List of that dataclass.
    TODO: Multigateway configuration will pull multiple jobs

    Args:
        db: A configured firestore client instance.
        tenant_id: The tenant_id to donwload information from.

    # TODO: Delete this function. Use get_current_project_model instead

    """
    configuration = pull_current_configuration_from_firestore(db, tenant_id)

    machines_groups_documents = (
        _pull_docs_from_collection_associated_with_configuration(
            db, machines_group_collection_name, configuration.reference
        )
    )

    if not machines_groups_documents:
        log.warning(
            "No Machines Groups documents were found associated "
            f"with configuration {configuration.to_dict()['configuration_id']}"
        )
        return None

    machines_documents = _pull_docs_from_collection_associated_with_configuration(
        db, machines_collection_name, configuration.reference
    )

    if not machines_documents:
        log.warning(
            f"No Machines documents were found associated "
            f"with configuration {configuration.to_dict()['configuration_id']}"
        )
        return None

    gateways_documents = _pull_docs_from_collection_associated_with_configuration(
        db, gateways_collection_name, configuration.reference
    )

    if not gateways_documents:
        log.error(
            "No Gateways documents were found associated "
            f"with configuration {configuration.to_dict()['configuration_id']}"
        )
        return None

    edges_documents = _pull_docs_from_collection_associated_with_configuration(
        db, edges_collection_name, configuration.reference
    )

    if not edges_documents:
        log.error(
            "No Edges documents were found associated "
            f"with configuration {configuration.to_dict()['configuration_id']}"
        )
        return None

    devices_documents = _pull_docs_from_collection_associated_with_configuration(
        db, devices_collection_name, configuration.reference
    )

    if not devices_documents:
        log.error(
            "No Devices documents were found associated "
            f"with configuration {configuration.to_dict()['configuration_id']}"
        )
        return None

    job_document = _pull_docs_from_collection_associated_with_configuration(
        db, jobs_collections_name, configuration.reference
    )

    if not job_document:
        log.error(
            "No Job document was found associated with "
            f"with configuration {configuration.to_dict()['configuration_id']}"
        )
        return None

    job_document = job_document[0]  # TODO: multigateway configuration
    job_config: JobConfig = JobConfig.build_from_dict(job_document)

    if not job_config:
        log.error("The JobConfig obtained from Firestore could not be parsed")
        return None

    ret = {
        "last_configuration": configuration,
        "machines_groups": machines_groups_documents,
        "machines": machines_documents,
        "gateways": gateways_documents,
        "edges": edges_documents,
        "devices": devices_documents,
        "job": job_document,
        "query_data": {},
    }

    return ret


def get_current_project_model(
        db: firestore.Client,
        tenant_id: str
) -> ProjectModel:
    """Returns the current Project Model configuration from firestore.

    If machines_groups are provided, then it will filter out the data that
    do not belong to those machines groups. The reason for this quick & dirty
    approach is the decision to host multiple clients in the same GCP project.

    TODO: Eventually we'll update the firestore project model schema to implement a better
        approach.

    Args:
        db: the firestore database client
        machines_groups_filter: an optional machines groups list to filter by.
    """
    configuration = pull_current_configuration_from_firestore(db, tenant_id)

    machines_groups_documents = (
        _pull_docs_from_collection_associated_with_configuration(
            db, machines_group_collection_name, configuration.reference, as_dict=False
        )
    )

    if not machines_groups_documents:
        log.warning(
            "No Machines Groups documents were found associated "
            f"with configuration {configuration.to_dict()['configuration_id']}"
        )
        return None

    machines_documents = _pull_docs_from_collection_associated_with_configuration(
        db, machines_collection_name, configuration.reference, as_dict=False
    )

    if not machines_documents:
        log.warning(
            f"No Machines documents were found associated "
            f"with configuration {configuration.to_dict()['configuration_id']}"
        )
        return None

    gateways_documents = _pull_docs_from_collection_associated_with_configuration(
        db, gateways_collection_name, configuration.reference, as_dict=False
    )

    if not gateways_documents:
        log.error(
            "No Gateways documents were found associated "
            f"with configuration {configuration.to_dict()['configuration_id']}"
        )
        return None

    edges_documents = _pull_docs_from_collection_associated_with_configuration(
        db, edges_collection_name, configuration.reference, as_dict=False
    )

    if not edges_documents:
        log.error(
            "No Edges documents were found associated "
            f"with configuration {configuration.to_dict()['configuration_id']}"
        )
        return None

    devices_documents = _pull_docs_from_collection_associated_with_configuration(
        db, devices_collection_name, configuration.reference, as_dict=False
    )

    if not devices_documents:
        log.error(
            "No Devices documents were found associated "
            f"with configuration {configuration.to_dict()['configuration_id']}"
        )
        return None

    job_documents = _pull_docs_from_collection_associated_with_configuration(
        db, jobs_collections_name, configuration.reference, as_dict=False
    )

    if not job_documents:
        log.error(
            "No Job document was found associated with "
            f"with configuration {configuration.to_dict()['configuration_id']}"
        )
        return None

    return ProjectModel(
        configuration,
        machines_groups_documents,
        machines_documents,
        gateways_documents,
        edges_documents,
        devices_documents,
        job_documents,
    )

# ============================ Configuration Snapshot related methods =======================================
def _create_devices(
    db: firestore.Client,
    devices: List[Dict],
    configuration_ref: firestore.DocumentReference,
) -> Dict:
    d = {}
    for device_dict in devices:
        log.info(f'Creating device in Firestore: {device_dict["kind"]}')
        device_dict["configuration_ref"] = configuration_ref
        result = db.collection(devices_collection_name).add(device_dict)
        d[device_dict["kind"]] = result[1]
    return d


def _create_edges(
    db: firestore.Client,
    edges_dicts: List[Dict],
    devices_references: Dict[str, firestore.DocumentReference],
    configuration_reference: firestore.DocumentReference,
) -> Dict:
    d = {}
    for edge_dict in edges_dicts:
        log.info(f'Creating edge in Firestore: {edge_dict["logical_id"]}')
        edge_dict["configuration_ref"] = configuration_reference
        edge_dict["is_of_kind_ref"] = devices_references[edge_dict["device_kind"]]
        if "edge_parameters" in edge_dict:
            edge_dict["parameters"] = edge_dict.pop("edge_parameters")
        result = db.collection(edges_collection_name).add(edge_dict)
        d[edge_dict["logical_id"]] = result[1]

    return d


def _create_jobs(
    db: firestore.Client,
    gateway_id_to_job: Dict[str, Dict],
    configuration_ref: firestore.DocumentReference,
) -> Dict:
    d = {}
    for gateway_id, job in gateway_id_to_job.items():
        job["configuration_ref"] = configuration_ref
        result = db.collection(jobs_collections_name).add(job)
        d[gateway_id] = result[1]

    return d


def _create_machines(
    db: firestore.Client,
    machines: List[Dict],
    edges_refs: Dict[str, firestore.DocumentReference],
    configuration_ref: firestore.DocumentReference,
) -> Dict:
    d = {}
    for machine_dict in machines:
        # replace associated_with_edges with edges references
        edges_refs_for_machine = list(
            map(
                lambda edge_logical_id: edges_refs[edge_logical_id],
                machine_dict["edges_ids"],
            )
        )
        machine_dict["associated_with_edges"] = edges_refs_for_machine
        machine_dict.pop("edges_ids")

        machine_dict["configuration_ref"] = configuration_ref

        result = db.collection(machines_collection_name).add(machine_dict)
        d[machine_dict["machine_id"]] = result[1]

    return d


def _create_machines_groups(
    db: firestore.Client,
    machines_groups: List[Dict],
    machines_refs: Dict[str, firestore.DocumentReference],
    configuration_ref: Dict[str, firestore.DocumentReference],
) -> Dict:
    d = {}
    for machine_group_dict in machines_groups:
        machines_docs = list(
            map(
                lambda machine_id: machines_refs[machine_id],
                machine_group_dict["machines"],
            )
        )

        machine_group_with_machines_refs = {
            **machine_group_dict,
            **{"machines": machines_docs, "configuration_ref": configuration_ref},
        }

        result = db.collection(machines_group_collection_name).add(
            machine_group_with_machines_refs
        )
        d[machine_group_dict["name"]] = result[1]

    return d


def _create_gateways(
    db: firestore.Client,
    gateways: List[Dict],
    created_machines: Dict[str, firestore.DocumentReference],
    created_jobs: Dict[str, firestore.DocumentReference],
    configuration_ref: firestore.DocumentReference,
) -> Dict:
    d = {}
    for gateway_dict in gateways:
        gateway_dict["configuration_ref"] = configuration_ref
        gateway_dict["has_job"] = created_jobs[gateway_dict["gateway_id"]]
        gateway_dict["machines"] = list(
            map(lambda machine: created_machines[machine], gateway_dict["machines_ids"])
        )  # list(created_machines.values())
        result = db.collection(gateways_collection_name).add(gateway_dict)
        d[gateway_dict["gateway_id"]] = result[1]

    return d


def create_configuration_snapshot(
    db: firestore.Client,
    project_model: ProjectModel,
    snapshot_type: ConfigurationSnapshotType,
    tenant_id: str
) -> bool:
    """Creates a new configuration snapshot for the provided project_model

    Returns:
        True if the configuration was created successfully.
    """
    # Pull the tenant document reference to index the configuration
    # Pull the tenant Document
    tenant = db.collection(tenants_collection).where(
        'tenant_id', '==', tenant_id
    ).get()

    if not tenant:
        log.error(f"Tenant document was not found for {tenant_id}")
        return False

    if len(tenant) > 1:
        log.warning(f"Multiple Tenants in the database with id: {tenant_id}.")

    tenant_ref = tenant[0].reference

    #  Creates new configuration
    new_configuration_version: str = (
        txp_config_version.get_next_normal_version(project_model.configuration_version)
        if snapshot_type == ConfigurationSnapshotType.NORMAL
        else txp_config_version.get_next_preliminary_version(
            project_model.configuration_version
        )
    )
    config_timestamp = datetime.datetime.now(
        pytz.timezone(settings.gateway.timezone)
    ).replace(microsecond=0)
    _, new_configuration = db.collection(configurations_collection_name).add(
        {
            "configuration_id": new_configuration_version,
            "since": config_timestamp,
            "server_timestamp": firestore.SERVER_TIMESTAMP,  # TODO: remove this filed if nobody is using it
            "tenant_ref": tenant_ref
        }
    )

    created_devices = {}
    created_edges = {}
    created_jobs = {}
    created_machines = {}
    created_machines_group = {}
    created_gateways = {}

    try:
        created_devices = _create_devices(
            db, list(project_model.device_kinds_table.values()), new_configuration
        )
        log.info("New devices documents created on Snapshot")

        edges = list(
            map(
                lambda edge_descriptor: json.loads(
                    json.dumps(edge_descriptor, cls=ComplexEncoder)
                ),
                list(project_model.edges_table.values()),
            )
        )
        created_edges = _create_edges(db, edges, created_devices, new_configuration)
        log.info("New edges documents created on Snapshot")

        jobs = dict(
            map(
                lambda dict_entry: (
                    dict_entry[0],
                    json.loads(json.dumps(dict_entry[1], cls=ComplexEncoder)),
                ),
                project_model.jobs_table_by_gateway.items(),
            )
        )
        created_jobs = _create_jobs(db, jobs, new_configuration)
        log.info("New jobs documents created on Snapshot")

        machines = list(
            map(
                lambda machine_metadata: json.loads(
                    json.dumps(machine_metadata, cls=ComplexEncoder)
                ),
                list(project_model.machines_table.values()),
            )
        )
        created_machines = _create_machines(
            db, machines, created_edges, new_configuration
        )

        created_machines_group = _create_machines_groups(
            db,
            list(project_model.machines_groups_table.values()),
            created_machines,
            new_configuration,
        )

        gateways = list(
            map(
                lambda gateway_config: json.loads(
                    json.dumps(gateway_config, cls=ComplexEncoder)
                ),
                list(project_model.gateways_table.values()),
            )
        )
        created_gateways = _create_gateways(
            db, gateways, created_machines, created_jobs, new_configuration
        )

    except Exception as e:
        log.error(f"Unknown error while creating documents for new configuration: {e}")
        log.error(traceback.format_exc())
        log.warning(f"Reverting the changes...")
        documents = (
            list(created_devices.values())
            + list(created_edges.values())
            + list(created_jobs.values())
            + list(created_machines.values())
            + list(created_machines_group.values())
            + list(created_gateways.values())
        )
        for document_reference in documents:
            document_reference.delete()
        new_configuration.delete()
        return False

    else:
        log.info("All the new entities were created for the new configuration snapshot")
        return True



"""
This module implements the Computer Vision Analytics View for the application.

NOTE: This is just a dummy implementation of a MainView in order to shown the
mechanism provided by the `core_components` package.
"""
import pytz
from txp.common.config import settings
import txp.cloud.controllers.bigquery as bigquery
from pylab import *
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
from plotly.subplots import make_subplots
from txp.common.ml.annotation import AnnotationLabel
from txp.cloud.controllers.authentication import AuthenticationInfo

log = logging.getLogger(__name__)
log.setLevel(settings.txp.general_log_level)


def get_images_and_dates(signals_from_interval):
    if signals_from_interval is not None:
        images = []
        dates = []
        labels = []
        start = time.perf_counter()
        for signal in signals_from_interval:
            ts = signal["observation_timestamp"]
            utc = pytz.timezone("UTC")
            timezone = pytz.timezone("America/Mexico_City")
            date_time = pd.to_datetime(int(ts))
            localized_timestamp = utc.localize(date_time)
            new_timezone = localized_timestamp.astimezone(timezone)
            image = bigquery.get_image_from_raw_data(signal["data"])
            if image is not None:
                images.append(image)
                try:
                    labels_data = AnnotationLabel.parse_raw(signal["label"])
                    labels.append(f"{labels_data.label_value}")
                    dates.append(new_timezone.strftime("%H:%M:%S"))
                except:
                    dates.append(new_timezone.strftime("%H:%M:%S"))
                    labels.append(None)
            else:
                pass
        end = time.perf_counter()
        logging.info(
            f"La conversión total de las {len(images)} imagenes tardó {(end - start)}"
        )
        if not images:
            return None, None, None
        else:
            return images, dates, labels
    else:
        logging.info(f"No hay data disponible")
        return None, None, None


def get_animation(user, images, name, step_names, data_format, labels):
    """
    from PIL images shows a plotly fig.
    """
    imgs = []
    dates = []
    annotations = []
    fig = make_subplots(rows=1, cols=1)
    traces_added = False
    img_width = 640
    img_height = 480
    start = time.perf_counter()
    frames = []

    for img, date, label in list(zip(images, step_names, labels)):
        try:
            plot_fig = px.imshow(np.asarray(img))
            imgs.append(plot_fig)
            if not traces_added:
                fig.add_trace(imgs[0].data[0], row=1, col=1)
                traces_added = True
            dates.append(date)
            annotations.append(label)
            if user == AuthenticationInfo.UserRole.Admin:
                if label is not None:
                    layout_i = go.Layout(
                        annotations=[
                            go.layout.Annotation(
                                x=300,
                                y=50,
                                text=label,
                                showarrow=False,
                                font_size=20,
                                font_color="black",
                                bgcolor="white",
                                opacity=0.85,
                            )
                        ]
                    )
                else:
                    layout_i = go.Layout(
                        annotations=[
                            go.layout.Annotation(
                                x=300,
                                y=50,
                                text="Sin etiqueta",
                                showarrow=False,
                                font_size=20,
                                font_color="black",
                                bgcolor="white",
                                opacity=0.85,
                            )
                        ]
                    )

                frames.append(
                    go.Frame(
                        name=str(date),
                        data=[
                            plot_fig.data[0],
                        ],
                        layout=layout_i,
                    )
                )
            else:
                frames.append(
                    go.Frame(
                        name=str(date),
                        data=[
                            plot_fig.data[0],
                        ],
                    )
                )



        except Exception as e:
            log.error(f"Error catched: {e}. Probably a truncated image was found.")

    fig.frames = frames

    fig.update_layout(
        sliders=[
            {
                "active": 0,
                "x": 0.1,
                "y": 0,
                "len": 0.9,
                "pad": {"b": 10, "t": 40, "r": 10, "l": 10},
                "steps": [
                    {
                        "args": [
                            [fr.name],
                            {
                                "frame": {"duration": 0, "redraw": True},
                                "mode": "immediate",
                                "fromcurrent": True,
                            },
                        ],
                        "label": fr.name,
                        "method": "animate",
                    }
                    for fr in fig.frames
                ],
            }
        ],
    )
    fig.update_layout(
        updatemenus=[
            {
                "buttons": [
                    {
                        "args": [
                            None,
                            {
                                "frame": {"duration": 500, "redraw": True},
                                "fromcurrent": True,
                                "transition": {
                                    "duration": 300,
                                    "easing": "quadratic-in-out",
                                },
                            },
                        ],
                        "label": "▶▶",
                        "method": "animate",
                    },
                    {
                        "args": [
                            None,
                            {
                                "frame": {"duration": 1200, "redraw": True},
                                "fromcurrent": True,
                                "transition": {
                                    "duration": 200,
                                    "easing": "quadratic-in-out",
                                },
                            },
                        ],
                        "label": "▶",
                        "method": "animate",
                    },
                    {
                        "args": [
                            [None],
                            {
                                "frame": {"duration": 0, "redraw": True},
                                "mode": "immediate",
                                "transition": {"duration": 0},
                            },
                        ],
                        "label": "■",
                        "method": "animate",
                    },
                ],
                "direction": "left",
                "pad": {"r": 10, "t": 40},
                "showactive": False,
                "type": "buttons",
                "x": 0.1,
                "xanchor": "right",
                "y": 0,
                "yanchor": "top",
            }
        ]
    )
    fig.update_layout(
        height=img_height,
        width=img_width,
        margin=dict(l=10, r=10, b=10, t=10, pad=4),
    )
    # Configure axes
    fig.update_xaxes(visible=False)

    fig.update_yaxes(
        visible=False,
        # the scaleanchor attribute ensures that the aspect ratio stays constant
        scaleanchor="x",
    )
    end = time.perf_counter()
    logging.info(f"La construcción del grafico {data_format} tardó {(end - start)}")
    return fig


def get_images_analysis(
    user,
    normal_data,
    thermographic_data,
    range_time,
    data_format,
    images_data_format,
    timages_data_format,
    animation_data_format,
    images_charts
):
    normal_images, normal_dates, normal_labels = get_images_and_dates(normal_data)
    (
        thermographic_images,
        thermographic_dates,
        thermographic_labels,
    ) = get_images_and_dates(thermographic_data)

    if normal_images:
        images_charts.normal_imgs_chat = get_animation(
            user,
            normal_images,
            images_data_format,
            normal_dates,
            data_format,
            normal_labels,
        )
    else:
        images_charts.normal_imgs_chat = None

    log.info("Normal Images returned")

    if thermographic_images:
        images_charts.thermographic_imgs_datacharts = get_animation(
            user,
            thermographic_images,
            timages_data_format,
            thermographic_dates,
            data_format,
            thermographic_labels,
        )
    else:
        images_charts.thermographic_imgs_datacharts = None

    log.info("Thermographic Images returned")



import numpy as np
from common import iCOMOX_messages
from scipy.signal import welch
from scipy.signal.windows import hamming
from scipy.signal.filter_design import butter
from scipy.signal import sosfilt
import struct


class class_IM69D130:
    def __init__(self):
        # Initializing Fs_Mic, MFFT_Mic, Scaling_Mic, window
        Root_clk = 26e6
        CLKDIV = 10  # SPORT clock is determined by Root_clk/(2*(SPORT_DIV_xCLKDIV+1))
        self.Fs_Mic = Root_clk / (2*64*CLKDIV)  # 64 bits clock provide 2 samples: left & right. We further divide the SPORT clock by 2 to get 20.3125KSPS
        self.NFFT_Mic = 512
        self.Scaling_Mic = 10**(130/20)/32768   #2.8 * 1E2
        self.window = hamming(M=self.NFFT_Mic)
        # self.sos = butter(100, 300.0, 'hp', fs=self.Fs_Mic, output='sos') # 10th order high pass filter, with half BW @28Hz

    #def Mic_readings_to_SPL_units(self, Mic_samples):
        # remove any DC components and scale the samples to SPL units

        # For IM69D130 (16 MSBs out of 20 bits, as we currently truncates the 4 LSBs):
        # Y_AOP = 130 dBSPL = 3,162,277.66 SPL for X_AOP = 32768 LSB
        # Y0 = 0 SPL, for X0 = 0 LSB
        # Now we can fit a linear curve using 2 known points:
        # Y - Y0 = (X - X0)*(Y_AOP - Y0)/(X_AOP - X0)
        # Solving it results in the following linear curve:
        # Y[SPL] = 96.505[SPL/LSB] * X[LSB]
        #return self.Scaling_Mic * (np.array(Mic_samples) - np.mean(Mic_samples))
        #return self.Scaling_Mic * np.array(Mic_samples)

    def to_PSD(self, mic_units_spl):
        (freq_Acc, Pxxf_Mic) = welch(x=mic_units_spl, fs=self.Fs_Mic, window=self.window, noverlap=self.NFFT_Mic/2, nfft=self.NFFT_Mic, return_onesided=True, detrend=False)
        # Pxxf_Mic[0:2] = np.mean(Pxxf_Mic[2:])
        return freq_Acc, Pxxf_Mic

    def msg_bytes_to_samples_SPL(self, payload):
        unscaled_samples = struct.unpack("<{}h".format(iCOMOX_messages.IM69D130_SAMPLES_NUM), payload)
        # filtered = sosfilt(self.sos, unscaled_samples)
        # return self.Scaling_Mic * np.array(filtered)
        return self.Scaling_Mic * (np.array(unscaled_samples) - np.mean(unscaled_samples))  #- np.mean(unscaled_samples))
